@@ -81,7 +81,7 @@ let reserved =
 	(* these ones are defined in order to prevent recursion in some Std functions *)
 	["is";"as";"int";"uint";"const";"getTimer";"typeof";"parseInt";"parseFloat";
 	(* AS3 keywords which are not haXe ones *)
-	"each";"label";"finally";"with";"final";"internal";"native";"const";"namespace";"include";
+	"each";"label";"finally";"with";"final";"internal";"native";"const";"namespace";"include";"delete";
 	(* we don't include get+set since they are not 'real' keywords, but they can't be used as method names *)
 	];
 	h
@@ -181,6 +181,13 @@ let parent e =
 	| TParenthesis _ -> e
 	| _ -> mk (TParenthesis e) e.etype e.epos
 
+let default_value tstr =
+	match tstr with
+	| "int" | "uint" -> "0"
+	| "Number" -> "NaN"
+	| "Boolean" -> "false"
+	| _ -> "null"
+
 let rec type_str ctx t p =
 	match t with
 	| TEnum _ | TInst _ when List.memq t ctx.local_types ->
@@ -277,9 +284,11 @@ let gen_function_header ctx name f params p =
 	print ctx "function%s(" (match name with None -> "" | Some n -> " " ^ n);
 	concat ctx "," (fun (arg,c,t) ->
 		let arg = define_local ctx arg in
-		print ctx "%s : %s" arg (type_str ctx t p);
+		let tstr = type_str ctx t p in
+		print ctx "%s : %s" arg tstr;
 		match c with
-		| None -> ()
+		| None ->
+			if ctx.constructor_block then print ctx " = %s" (default_value tstr);
 		| Some c ->
 			spr ctx " = ";
 			gen_constant ctx p c
@@ -462,7 +471,7 @@ and gen_expr ctx e =
 		gen_value_op ctx e2;
 	| TField ({ eexpr = TTypeExpr t },s) when t_path t = ctx.curclass.cl_path && not (PMap.mem s ctx.locals) ->
 		print ctx "%s" (s_ident s)
-	| TField (e,s) ->
+	| TField (e,s) | TClosure (e,s) ->
    		gen_value ctx e;
 		gen_field_access ctx e.etype s
 	| TTypeExpr t ->
@@ -495,9 +504,12 @@ and gen_expr ctx e =
 		let b = save_locals ctx in
 		print ctx "{";
 		let bend = open_block ctx in
-		let cb = (if not ctx.constructor_block || not (Codegen.constructor_side_effects e) then
+		let cb = (if not ctx.constructor_block then
 			(fun () -> ())
-		else begin
+		else if not (Codegen.constructor_side_effects e) then begin
+			ctx.constructor_block <- false;
+			(fun () -> ())
+		end else begin
 			ctx.constructor_block <- false;
 			print ctx " if( !%s.skip_constructor ) {" (s_path ctx true (["flash"],"Boot") e.epos);
             (fun() -> print ctx "}")
@@ -603,6 +615,9 @@ and gen_expr ctx e =
 			b();
 		) catchs;
 	| TMatch (e,_,cases,def) ->
+		print ctx "{";
+		let bend = open_block ctx in
+		newline ctx;
 		let b = save_locals ctx in
 		let tmp = define_local ctx "$e" in
 		print ctx "var %s : enum = " tmp;
@@ -644,6 +659,9 @@ and gen_expr ctx e =
 			newline ctx;
 		);
 		spr ctx "}";
+		bend();
+		newline ctx;
+		spr ctx "}";
 		b()
 	| TSwitch (e,cases,def) ->
 		spr ctx "switch";
@@ -724,6 +742,7 @@ and gen_value ctx e =
 	| TArray _
 	| TBinop _
 	| TField _
+	| TClosure _
 	| TTypeExpr _
 	| TParenthesis _
 	| TObjectDecl _
@@ -806,7 +825,7 @@ let generate_field ctx static f =
 	let rights = (if static then "static " else "") ^ (if public then "public" else "protected") in
 	let p = ctx.curclass.cl_pos in
 	match f.cf_expr with
-	| Some { eexpr = TFunction fd } when f.cf_set = MethodCantAccess || f.cf_set = NeverAccess ->
+	| Some { eexpr = TFunction fd } when f.cf_set = MethodAccess false || f.cf_set = NeverAccess ->
 		print ctx "%s " rights;
 		let rec loop c =
 			match c.cl_super with
@@ -832,23 +851,19 @@ let generate_field ctx static f =
 				concat ctx "," (fun (arg,o,t) ->
 					let tstr = type_str ctx t p in
 					print ctx "%s : %s" arg tstr;
-					if o then print ctx " = %s" (match tstr with
-						| "int" | "uint" -> "0"
-						| "Number" -> "NaN"
-						| "Boolean" -> "false"
-						| _ -> "null");					
+					if o then print ctx " = %s" (default_value tstr);
 				) args;
 				print ctx ") : %s " (type_str ctx r p);
 			| _ -> ()
 		else
-		if (match f.cf_get with MethodAccess m -> true | _ -> match f.cf_set with MethodAccess m -> true | _ -> false) then begin
+		if (match f.cf_get with CallAccess m -> true | _ -> match f.cf_set with CallAccess m -> true | _ -> false) then begin
 			let t = type_str ctx f.cf_type p in
 			let id = s_ident f.cf_name in
 			(match f.cf_get with
 			| NormalAccess ->
 				print ctx "%s function get %s() : %s { return $%s; }" rights id t id;
 				newline ctx
-			| MethodAccess m ->
+			| CallAccess m ->
 				print ctx "%s function get %s() : %s { return %s(); }" rights id t m;
 				newline ctx
 			| _ -> ());
@@ -856,7 +871,7 @@ let generate_field ctx static f =
 			| NormalAccess ->
 				print ctx "%s function set %s( __v : %s ) : void { $%s = __v; }" rights id t id;
 				newline ctx
-			| MethodAccess m ->
+			| CallAccess m ->
 				print ctx "%s function set %s( __v : %s ) : void { %s(__v); }" rights id t m;
 				newline ctx
 			| _ -> ());
@@ -875,8 +890,8 @@ let rec define_getset ctx stat c =
 		Hashtbl.add ctx.get_sets (name,stat) f.cf_name
 	in
 	let field f =
-		(match f.cf_get with MethodAccess m -> def f m | _ -> ());
-		(match f.cf_set with MethodAccess m -> def f m | _ -> ())
+		(match f.cf_get with CallAccess m -> def f m | _ -> ());
+		(match f.cf_set with CallAccess m -> def f m | _ -> ())
 	in
 	List.iter field (if stat then c.cl_ordered_statics else c.cl_ordered_fields);
 	match c.cl_super with
@@ -907,7 +922,7 @@ let generate_class ctx c =
 		let f = { f with
 			cf_name = snd c.cl_path;
 			cf_public = true;
-			cf_set = MethodCantAccess;
+			cf_set = MethodAccess false;
 		} in
 		ctx.constructor_block <- true;
 		generate_field ctx false f;
@@ -974,6 +989,8 @@ let generate_enum ctx e =
 		| _ ->
 			print ctx "public static var %s : %s = new %s(\"%s\",%d)" c.ef_name ename ename c.ef_name c.ef_index;
 	) e.e_constrs;
+	newline ctx;
+	print ctx "public static var __constructs__ : Array = [%s];" (String.concat "," (List.map (fun s -> "\"" ^ Ast.s_escape s ^ "\"") e.e_names));
 	cl();
 	newline ctx;
 	print ctx "}";
