@@ -1,12 +1,3 @@
-(*
-TODO
-- check for static method name clashes
-- debug version
-- runtime check for undefined fields
-OPTIMIZATION
-- replace eval for statements with functions/inlines
-- replace closures (eval) with functions
-*)
 (*
  *  haXe/PHP Compiler
  *  Copyright (c)2008 Franco Ponticelli
@@ -40,8 +31,9 @@ type context = {
 	ch : out_channel;
 	buf : Buffer.t;
 	path : path;
-	mutable get_sets : (string * bool,string) Hashtbl.t;
+	stack : Codegen.stack_context;
 	mutable curclass : tclass;
+	mutable curmethod : string;
 	mutable tabs : string;
 	mutable in_value : string option;
 	mutable in_loop : bool;
@@ -61,6 +53,106 @@ type context = {
 	mutable cwd : string;
 }
 
+let join_class_path path separator =
+	let result = match fst path, snd path with
+	| [], s -> s
+	| el, s -> String.concat separator el ^ separator ^ s in
+	if (String.contains result '+') then begin
+		let idx = String.index result '+' in
+		(String.sub result 0 idx) ^ (String.sub result (idx+1) ((String.length result) - idx -1 ) )
+	end else
+		result;;
+
+(*  Get a string to represent a type.
+	 The "suffix" will be nothing or "_obj", depending if we want the name of the
+	 pointer class or the pointee (_obj class *)
+let rec class_string klass suffix params =
+	(match klass.cl_path with
+	(* Array class *)
+	|  ([],"Array") -> (snd klass.cl_path) ^ suffix ^ "<" ^ (String.concat ","
+					 (List.map type_string  params) ) ^ " >"
+	| _ when klass.cl_kind=KTypeParameter -> "Dynamic"
+	|  ([],"#Int") -> "/* # */int"
+	|  (["haxe";"io"],"Unsigned_char__") -> "unsigned char"
+	|  ([],"Class") -> "Class"
+	|  ([],"Null") -> (match params with
+			| [t] ->
+				(match follow t with
+				| TInst ({ cl_path = [],"Int" },_)
+				| TInst ({ cl_path = [],"Float" },_)
+				| TEnum ({ e_path = [],"Bool" },_) -> "Dynamic"
+				| _ -> "/*NULL*/" ^ (type_string t) )
+			| _ -> assert false); 
+	(* Normal class *)
+	| _ -> (join_class_path klass.cl_path "::") ^ suffix
+	)
+and type_string_suff suffix haxe_type =
+	(match haxe_type with
+	| TMono r -> (match !r with None -> "Dynamic" | Some t -> type_string_suff suffix t)
+	| TEnum ({ e_path = ([],"Void") },[]) -> "Void"
+	| TEnum ({ e_path = ([],"Bool") },[]) -> "bool"
+	| TInst ({ cl_path = ([],"Float") },[]) -> "double"
+	| TInst ({ cl_path = ([],"Int") },[]) -> "int"
+	| TEnum (enum,params) ->  (join_class_path enum.e_path "::") ^ suffix
+	| TInst (klass,params) ->  (class_string klass suffix params)
+	| TType (type_def,params) ->
+		(match type_def.t_path with
+		| [] , "Null" ->
+			(match params with
+			| [t] ->
+				(match follow t with
+				| TInst ({ cl_path = [],"Int" },_)
+				| TInst ({ cl_path = [],"Float" },_)
+				| TEnum ({ e_path = [],"Bool" },_) -> "Dynamic"
+				| _ -> type_string_suff suffix t)
+			| _ -> assert false);
+		| [] , "Array" ->
+			(match params with
+			| [t] -> "Array<" ^ (type_string (follow t) ) ^ " >"
+			| _ -> assert false)
+		| _ ->  type_string_suff suffix (apply_params type_def.t_types params type_def.t_type)
+		)
+	| TFun (args,haxe_type) -> "Dynamic"
+	| TAnon anon -> "Dynamic"
+	| TDynamic haxe_type -> "Dynamic"
+	| TLazy func -> type_string_suff suffix ((!func)())
+	)
+and type_string haxe_type = 
+	type_string_suff "" haxe_type;;
+
+let debug_expression expression type_too =
+	"/* " ^
+	(match expression.eexpr with
+	| TConst _ -> "TConst"
+	| TLocal _ -> "TLocal"
+	| TEnumField _ -> "TEnumField"
+	| TArray (_,_) -> "TArray"
+	| TBinop (_,_,_) -> "TBinop"
+	| TField (_,_) -> "TField"
+	| TClosure _ -> "TClosure"
+	| TTypeExpr _ -> "TTypeExpr"
+	| TParenthesis _ -> "TParenthesis"
+	| TObjectDecl _ -> "TObjectDecl"
+	| TArrayDecl _ -> "TArrayDecl"
+	| TCall (_,_) -> "TCall"
+	| TNew (_,_,_) -> "TNew"
+	| TUnop (_,_,_) -> "TUnop"
+	| TFunction _ -> "TFunction"
+	| TVars _ -> "TVars"
+	| TBlock _ -> "TBlock"
+	| TFor (_,_,_,_) -> "TFor"
+	| TIf (_,_,_) -> "TIf"
+	| TWhile (_,_,_) -> "TWhile"
+	| TSwitch (_,_,_) -> "TSwitch"
+	| TMatch (_,_,_,_) -> "TMatch"
+	| TTry (_,_) -> "TTry"
+	| TReturn _ -> "TReturn"
+	| TBreak -> "TBreak"
+	| TContinue -> "TContinue"
+	| TThrow _ -> "TThrow" ) ^
+	(if (type_too) then " = " ^ (type_string expression.etype) else "") ^
+	" */";;
+
 let rec escphp n =
 	if n = 0 then "" else if n = 1 then "\\" else ("\\\\" ^ escphp (n-1))
 
@@ -76,6 +168,7 @@ let s_expr_expr e =
 	| TArray (_,_) -> "TArray"
 	| TBinop (_,_,_) -> "TBinop"
 	| TField (_,_) -> "TField"
+	| TClosure (_,_) -> "TClosure"
 	| TTypeExpr _ -> "TTypeExpr"
 	| TParenthesis _ -> "TParenthesis"
 	| TObjectDecl _ -> "TObjectDecl"
@@ -189,14 +282,21 @@ haxe reserved words that match php ones: break, case, class, continue, default,
  *)
 (* PHP only (for future use): cfunction, old_function *)
 	match n with
-	| "and" | "or" | "xor" | "__FILE__" | "exception" (* PHP 5 *) | "__LINE__" | "array"
+	| "and" | "or" | "xor" | "__FILE__" | "exception" | "__LINE__" | "array"
 	| "as" | "const" | "declare" | "die" | "echo"| "elseif" | "empty"
 	| "enddeclare" | "endfor" | "endforeach" | "endif" | "endswitch"
 	| "endwhile" | "eval" | "exit" | "foreach"| "global" | "include"
 	| "include_once" | "isset" | "list" | "print" | "require" | "require_once"
-	| "unset" | "use" | "__FUNCTION__" | "__CLASS__" | "__METHOD__" | "final" (* PHP 5 *)
-	| "php_user_filter" (* PHP 5 *) | "protected" (* PHP 5 *) | "abstract" (* PHP 5 *)
-	| "clone" (* PHP 5 *) -> suf ^ n
+	| "unset" | "use" | "__FUNCTION__" | "__CLASS__" | "__METHOD__" | "final" 
+	| "php_user_filter" | "protected" | "abstract" | "__set" | "__get" | "__call"
+	| "clone" -> suf ^ n
+	| _ -> n
+	
+let s_ident_local n =
+	let suf = "h" in
+	match n with
+	| "GLOBALS" | "_SERVER" | "_GET" | "_POST" | "_COOKIE" | "_FILES" 
+	| "_ENV" | "_REQUEST" | "_SESSION" -> suf ^ n
 	| _ -> n
 
 let write_resource dir name data =
@@ -211,6 +311,9 @@ let write_resource dir name data =
 	let ch = open_out_bin (rdir ^ "/" ^ name) in
 	output_string ch data;
 	close_out ch
+	
+let stack_init com use_add =
+	Codegen.stack_context_init com "GLOBALS['%s']" "GLOBALS['%e']" "spos" "tmp" use_add null_pos
 
 let init com cwd path def_type =
 	let rec create acc = function
@@ -231,6 +334,7 @@ let init com cwd path def_type =
 	Hashtbl.add imports (snd path) [fst path];
 	{
 		com = com;
+		stack = stack_init com false;
 		tabs = "";
 		ch = ch;
 		path = path;
@@ -242,11 +346,11 @@ let init com cwd path def_type =
 		extern_required_paths = [];
 		extern_classes_with_init = [];
 		curclass = null_class;
+		curmethod = "";
 		locals = PMap.empty;
 		inv_locals = PMap.empty;
 		local_types = [];
 		inits = [];
-		get_sets = Hashtbl.create 0;
 		constructor_block = false;
 		quotes = 0;
 		dynamic_methods = [];
@@ -306,7 +410,7 @@ let save_locals ctx =
 
 let define_local ctx l =
 	let rec loop n =
-	let name = (if n = 1 then s_ident l else l ^ string_of_int n) in
+	let name = (if n = 1 then s_ident_local l else s_ident_local (l ^ string_of_int n)) in
 	if PMap.mem name ctx.inv_locals then
 		loop (n+1)
 	else begin
@@ -346,12 +450,12 @@ let handle_break ctx e =
 				ctx.handle_break <- snd old;
 				newline ctx;
 				let p = escphp ctx.quotes in
-				print ctx "} catch(Exception %s$e) { if( %s$e->getMessage() != %s\"__break__%s\" ) throw %s$e; }" p p p p p;
+				print ctx "} catch(_hx_break_exception %s$e){}" p;
 			)
 
 let this ctx =
 	let p = escphp ctx.quotes in
-	if ctx.in_value <> None then (p ^ "$__this") else (p ^ "$this")
+	if ctx.in_value <> None then (p ^ "$this") else (p ^ "$this")
 
 let escape_bin s quotes =
 	let b = Buffer.create 0 in
@@ -379,6 +483,8 @@ let gen_constant ctx p = function
 
 let s_funarg ctx arg t p c =
 	let byref = if (String.length arg > 7 && String.sub arg 0 7 = "byref__") then "&" else "" in
+	print ctx "%s%s$%s" byref (escphp ctx.quotes) (s_ident_local arg)
+(*
 	(match t with
 	| TInst (cl,_) ->
 		(match cl.cl_path with
@@ -399,6 +505,7 @@ let s_funarg ctx arg t p c =
 	| _ ->
 		print ctx "%s%s$%s" byref (escphp ctx.quotes) arg;
 	)
+*)
 
 let is_in_dynamic_methods ctx e s =
 	List.exists (fun dm ->
@@ -407,13 +514,11 @@ let is_in_dynamic_methods ctx e s =
 	) ctx.all_dynamic_methods
 
 let is_dynamic_method f =
-	match follow f.cf_type with
-	| TFun _ when f.cf_expr = None -> true
-	| _ ->
-		(match f.cf_expr with
-		| Some { eexpr = TFunction fd } -> f.cf_set = NormalAccess
+	(match f.cf_set with
+		| NormalAccess -> true
+		| MethodAccess true -> true
 		| _ -> false)
-
+		
 let fun_block ctx f p =
 	let e = (match f.tf_expr with { eexpr = TBlock [{ eexpr = TBlock _ } as e] } -> e | e -> e) in
 	let e = List.fold_left (fun e (a,c,t) ->
@@ -421,6 +526,9 @@ let fun_block ctx f p =
 		| None | Some TNull -> e
 		| Some c -> Codegen.concat (Codegen.set_default ctx.com a c t p) e
 	) e f.tf_args in
+	if ctx.com.debug then begin
+		Codegen.stack_block ctx.stack ctx.curclass ctx.curmethod e
+	end else 
 		mk_block e
 
 let gen_function_header ctx name f params p =
@@ -554,6 +662,17 @@ and gen_call ctx e el =
 		spr ctx ", array(";
 		concat ctx ", " (gen_value ctx) el;
 		spr ctx "))"
+(*
+	| TField (ex,name), el ->
+		spr ctx (debug_expression ex true);
+		ctx.is_call <- true;
+		spr ctx "call_user_func_array(";
+		gen_value ctx e;
+		ctx.is_call <- false;
+		spr ctx ", array(";
+		concat ctx ", " (gen_value ctx) el;
+		spr ctx "))"
+*)
 	| _ ->
 		ctx.is_call <- true;
 		gen_value ctx e;
@@ -855,7 +974,8 @@ and gen_expr ctx e =
 	| TConst c ->
 		gen_constant ctx e.epos c
 	| TLocal s ->
-		spr ctx ((escphp ctx.quotes) ^ "$" ^ (try PMap.find s ctx.locals with Not_found -> error ("Unknown local " ^ s) e.epos))
+		spr ctx ((escphp ctx.quotes) ^ "$" ^ (try PMap.find s ctx.locals with Not_found -> (s_ident_local s)))
+(*		spr ctx ((escphp ctx.quotes) ^ "$" ^ (s_ident_local s)) *)
 	| TEnumField (en,s) ->
 		(match (try PMap.find s en.e_constrs with Not_found -> error ("Unknown local " ^ s) e.epos).ef_type with
 		| TFun (args,_) -> print ctx "%s::%s" (s_path ctx en.e_path en.e_extern e.epos) (s_ident s)
@@ -878,7 +998,7 @@ and gen_expr ctx e =
 			(match e.eexpr with
 			| TArray(te1, te2) ->
 				gen_value ctx te1;
-				spr ctx "->__a[";
+				spr ctx "->a[";
 				gen_value ctx te2;
 				spr ctx "]";
 			| _ ->
@@ -887,7 +1007,7 @@ and gen_expr ctx e =
 			(match e.eexpr with
 			| TArray(te1, te2) ->
 				gen_value ctx te1;
-				spr ctx "->__a[";
+				spr ctx "->a[";
 				gen_value ctx te2;
 				spr ctx "]";
 			| TField (e1,s) ->
@@ -913,14 +1033,17 @@ and gen_expr ctx e =
 				gen_value_op ctx e2;
 				spr ctx ")";
 			| _ ->
-(*
-				gen_field_op ctx e1;
-*)
-
 				leftsidef e1;
 				spr ctx " = ";
-				
 				gen_value_op ctx e2;)
+		| Ast.OpAssignOp(Ast.OpAdd) when (is_uncertain_expr e1 && is_uncertain_expr e2) ->
+			leftside e1;
+			spr ctx " = ";
+			spr ctx "_hx_add(";
+			gen_value_op ctx e1;
+			spr ctx ", ";
+			gen_value_op ctx e2;
+			spr ctx ")";
 		| Ast.OpAssignOp(Ast.OpAdd) when (is_string_expr e1 || is_string_expr e2) ->
 			leftside e1;
 			spr ctx " .= ";
@@ -941,6 +1064,12 @@ and gen_expr ctx e =
 			leftsidec e1;
 			print ctx " %s " (Ast.s_binop op);
 			gen_value_op ctx e2;
+		| Ast.OpAdd when (is_uncertain_expr e1 && is_uncertain_expr e2) ->
+			spr ctx "_hx_add(";
+			gen_value_op ctx e1;
+			spr ctx ", ";
+			gen_value_op ctx e2;
+			spr ctx ")";
 		| Ast.OpAdd when (is_string_expr e1 || is_string_expr e2) ->
 			gen_value_op ctx e1;
 			spr ctx " . ";
@@ -1034,14 +1163,14 @@ and gen_expr ctx e =
 			print ctx " %s " (Ast.s_binop op);
 			gen_value_op ctx e2;
 		);
-	| TField (e1,s) ->
+	| TField (e1,s) | TClosure (e1,s) ->
 		(match follow e.etype with
 		| TFun (args, _) ->
 			let p = escphp ctx.quotes in
 			(if ctx.is_call then begin
 				gen_field_access ctx false e1 s
 	  		end else if is_in_dynamic_methods ctx e1 s then begin
-	  			gen_field_access ctx true e1 s
+	  			gen_field_access ctx true e1 s;
 	  		end else begin
 				let ob ex = 
 					(match ex with
@@ -1096,8 +1225,7 @@ and gen_expr ctx e =
 			);
 	| TBreak ->
 		if not ctx.in_loop then unsupported e.epos;
-		let p = escphp ctx.quotes in
-		if ctx.handle_break then print ctx "throw new Exception(%s\"__break__%s\")" p p else spr ctx "break"
+		if ctx.handle_break then spr ctx "throw new _hx_break_exception()" else spr ctx "break"
 	| TContinue ->
 		if not ctx.in_loop then unsupported e.epos;
 		spr ctx "continue"
@@ -1112,34 +1240,37 @@ and gen_expr ctx e =
 				(fun () -> ())
 			else begin
 				ctx.constructor_block <- false;
-			if List.length ctx.dynamic_methods > 0 then newline ctx else spr ctx " ";
-			List.iter (fun (f) ->
-				let name = f.cf_name in
-				match f.cf_expr with
-				| Some { eexpr = TFunction fd } ->
-					print ctx "$this->%s = array(new _hx_lambda(array(), $this, array(" name;
-					let cargs = ref 0 in
-					concat ctx "," (fun (arg,o,t) ->
-						let arg = define_local ctx arg in
-						print ctx "'%s'" arg;
-						incr cargs;
-					) fd.tf_args;
-					print ctx "), \"";
-					let old = ctx.in_value in
-					ctx.in_value <- Some name;
-					ctx.quotes <- ctx.quotes + 1;
-					gen_expr ctx (fun_block ctx fd e.epos);
-					ctx.quotes <- ctx.quotes - 1;
-					ctx.in_value <- old;
-					print ctx "\"), 'execute%d')" !cargs;
-					newline ctx;
-				| _ -> ()
-			) ctx.dynamic_methods;
-			if Codegen.constructor_side_effects e then begin
-				print ctx "if( !%s::$skip_constructor ) {" (s_path ctx (["php"],"Boot") false e.epos);
-				(fun() -> print ctx "}")
-			end else
-				(fun() -> ())
+				if List.length ctx.dynamic_methods > 0 then newline ctx else spr ctx " ";
+				List.iter (fun (f) ->
+					let name = f.cf_name in
+					match f.cf_expr with
+					| Some { eexpr = TFunction fd } ->
+						print ctx "if(!isset($this->%s)) $this->%s = array(new _hx_lambda(array(), $this, array(" name name;
+						let cargs = ref 0 in
+						concat ctx "," (fun (arg,o,t) ->
+							let arg = define_local ctx arg in
+							print ctx "'%s'" arg;
+							incr cargs;
+						) fd.tf_args;
+						print ctx "), \"";
+						let old = ctx.in_value in
+						ctx.in_value <- Some name;
+						ctx.quotes <- ctx.quotes + 1;
+						gen_expr ctx (fun_block ctx fd e.epos);
+						ctx.quotes <- ctx.quotes - 1;
+						ctx.in_value <- old;
+						print ctx "\"), 'execute%d')" !cargs;
+						newline ctx;
+					| _ -> ()
+				) ctx.dynamic_methods;
+				if Codegen.constructor_side_effects e then begin
+					print ctx "if( !%s::$skip_constructor ) {" (s_path ctx (["php"],"Boot") false e.epos);
+					(fun() -> print ctx "}";
+					
+					
+					)
+				end else
+					(fun() -> ());
 			end) in
 		List.iter (fun e -> newline ctx; gen_expr ctx e) el;
 		bend();
@@ -1149,9 +1280,12 @@ and gen_expr ctx e =
 		b();
 	| TFunction f ->
 		let old = ctx.in_value, ctx.in_loop in
+		let old_meth = ctx.curmethod in
 		ctx.in_value <- None;
 		ctx.in_loop <- false;
+		ctx.curmethod <- ctx.curmethod ^ "@" ^ string_of_int (Lexer.find_line_index ctx.com.lines e.epos);
 		gen_inline_function ctx f [] e.epos;
+		ctx.curmethod <- old_meth;
 		ctx.in_value <- fst old;
 		ctx.in_loop <- snd old;
 	| TCall (ec,el) ->
@@ -1166,16 +1300,8 @@ and gen_expr ctx e =
 			gen_string_static_call ctx s ef el
 		| TField (ef,s) when is_string_expr ef ->
 			gen_string_call ctx s ef el
-(*		| TField (ef,s) when is_anonym_expr ef ->
-			if could_be_string_call s then begin
-				gen_uncertain_string_call ctx s ef el
-			end else
-				gen_call ctx ec el
-*)
 		| TField (ef,s) when is_anonym_expr ef && could_be_string_call s ->
 			gen_uncertain_string_call ctx s ef el
-(*		| TCall _ ->
-			gen_call ctx ec el *)
 		| _ ->
 			gen_call ctx ec el);
 	| TArrayDecl el ->
@@ -1194,9 +1320,9 @@ and gen_expr ctx e =
 			let n = define_local ctx n in
 			match v with
 			| None -> 
-				print ctx "%s = null" n
+				print ctx "%s = null" (s_ident_local n)
 			| Some e ->
-				print ctx "%s = " n;
+				print ctx "%s = " (s_ident_local n);
 				gen_value ctx e
 		) vl;
 	| TNew (c,_,el) ->
@@ -1235,7 +1361,7 @@ and gen_expr ctx e =
 		(match e.eexpr with
 		| TArray(te1, te2) ->
 			gen_value ctx te1;
-			spr ctx "->__a[";
+			spr ctx "->a[";
 			gen_value ctx te2;
 			spr ctx "]";
 		| TField (e1,s) ->
@@ -1246,7 +1372,7 @@ and gen_expr ctx e =
 		(match e.eexpr with
 		| TArray(te1, te2) ->
 			gen_value ctx te1;
-			spr ctx "->__a[";
+			spr ctx "->a[";
 			gen_value ctx te2;
 			spr ctx "]";
 		| TField (e1,s) ->
@@ -1276,7 +1402,7 @@ and gen_expr ctx e =
 	| TFor (v,t,it,e) ->
 		let handle_break = handle_break ctx e in
 		let b = save_locals ctx in
-		let tmp = define_local ctx "__it__" in
+		let tmp = define_local ctx "it" in
 		let v = define_local ctx v in
 		let p = escphp ctx.quotes in
 		print ctx "%s$%s = " p tmp;
@@ -1294,7 +1420,7 @@ and gen_expr ctx e =
 	| TTry (e,catchs) ->
 		spr ctx "try ";
 		gen_expr ctx (mk_block e);
-		let ex = define_local ctx "__e__" in
+		let ex = define_local ctx "e" in
 		print ctx "catch(Exception %s$%s) {" (escphp ctx.quotes) ex;
 		let p = escphp ctx.quotes in
 		let first = ref true in
@@ -1342,7 +1468,7 @@ and gen_expr ctx e =
 			print ctx " else throw %s$%s; }" (escphp ctx.quotes) ex;
 	| TMatch (e,_,cases,def) ->
 		let b = save_locals ctx in
-		let tmp = define_local ctx "__t__" in
+		let tmp = define_local ctx "t" in
 		print ctx "%s$%s = " (escphp ctx.quotes) tmp;
 		gen_value ctx e;
 		newline ctx;
@@ -1410,14 +1536,14 @@ and gen_expr ctx e =
 and gen_value ctx e =
 	let assign e =
 		mk (TBinop (Ast.OpAssign,
-			mk (TLocal (match ctx.in_value with None -> assert false | Some v -> "__r__")) t_dynamic e.epos,
+			mk (TLocal (match ctx.in_value with None -> assert false | Some v -> "r")) t_dynamic e.epos,
 			e
 		)) e.etype e.epos
 	in
 	let value bl =
 		let old = ctx.in_value, ctx.in_loop in
 		let locs = save_locals ctx in
-		let tmp = define_local ctx "__r__" in
+		let tmp = define_local ctx "r" in
 		ctx.in_value <- Some tmp;
 		ctx.in_loop <- false;
 		let b =
@@ -1425,7 +1551,7 @@ and gen_value ctx e =
 			print ctx "eval(%s\"" (escphp ctx.quotes);
 			ctx.quotes <- (ctx.quotes + 1);
 			let p = (escphp ctx.quotes) in
-			print ctx "if(isset(%s$this)) %s$__this =& %s$this;" p p p;
+			print ctx "if(isset(%s$this)) %s$this =& %s$this;" p p p;
 			let b = open_block ctx in
 			b
 		end else
@@ -1453,6 +1579,7 @@ and gen_value ctx e =
 	| TArray _
 	| TBinop _
 	| TField _
+	| TClosure _
 	| TParenthesis _
 	| TObjectDecl _
 	| TArrayDecl _
@@ -1538,17 +1665,6 @@ let generate_self_method ctx rights m static setter =
 			print ctx "%s function %s() { return call_user_func($this->%s); }" rights (s_ident m) (s_ident m)
 	);
 	newline ctx
-
-let rec field_exists_in_hierarchy cl name =
-	if PMap.exists name cl.cl_fields then true
-	else (
-		match cl.cl_super with
-		| Some (c,_) -> 
-			field_exists_in_hierarchy c name
-		| None -> 
-			false
-	)
-	
 	
 let generate_field ctx static f =
 	newline ctx;
@@ -1558,10 +1674,14 @@ let generate_field ctx static f =
 	let p = ctx.curclass.cl_pos in
 	match f.cf_expr with
 	| Some { eexpr = TFunction fd } ->
-		if static && field_exists_in_hierarchy ctx.curclass f.cf_name then error ("Can't redeclare method (PHP limitation): " ^ f.cf_name) ctx.curclass.cl_pos;
+		if f.cf_name = "__construct" then
+			ctx.curmethod <- "new"
+		else
+			ctx.curmethod <- f.cf_name;
 		spr ctx (rights ^ " ");
 		(match f.cf_set with
-		| NormalAccess when (match fd.tf_expr.eexpr with | TBlock _ -> true | _ -> false) ->
+		| NormalAccess
+		| MethodAccess true ->
 			gen_dynamic_function ctx static (s_ident f.cf_name) fd f.cf_params p
 		| _ ->
 			gen_function ctx (s_ident f.cf_name) fd f.cf_params p
@@ -1578,7 +1698,7 @@ let generate_field ctx static f =
 			| _ -> spr ctx "//"; ()
 		else if
 			(match f.cf_get, f.cf_set with
-			| MethodAccess m1, MethodAccess m2 ->
+			| CallAccess m1, CallAccess m2 ->
 				if not (is_method_defined ctx m1 static) then (
 					generate_self_method ctx rights m1 static false;
 					print ctx "%s $%s" rights (s_ident m1);
@@ -1590,11 +1710,11 @@ let generate_field ctx static f =
 				if (is_method_defined ctx m1 static) && (is_method_defined ctx m2 static) then
 					spr ctx "//";
 				true
-			| MethodAccess m, _ ->
+			| CallAccess m, _ ->
 				if not (is_method_defined ctx m static) then generate_self_method ctx rights m static false;
 				print ctx "%s $%s" rights (s_ident f.cf_name);
 				true
-			| _, MethodAccess m ->
+			| _, CallAccess m ->
 				if not (is_method_defined ctx m static) then generate_self_method ctx rights m static true;
 				print ctx "%s $%s" rights (s_ident f.cf_name);
 				true
@@ -1602,7 +1722,22 @@ let generate_field ctx static f =
 				false) then
 				()
 		else begin
-			print ctx "%s $%s" rights (s_ident f.cf_name);
+			let name = s_ident f.cf_name in
+			if static then
+				(match f.cf_set with
+				| NormalAccess -> 
+					(match follow f.cf_type with
+					| TFun _
+					| TDynamic _ ->
+						print ctx "static function %s() { $args = func_get_args(); return call_user_func_array(self::$%s, $args); }" name name;
+						newline ctx;
+					| _ ->
+						()
+					)
+				| _ ->
+					()
+				);
+			print ctx "%s $%s" rights name;
 			match f.cf_expr with
 			| None -> ()
 			| Some e ->
@@ -1623,7 +1758,17 @@ let generate_static_field_assign ctx path f =
 			| TConst _ -> ()
 			| TFunction fd ->
 				(match f.cf_set with
-				| NormalAccess when (match fd.tf_expr.eexpr with | TBlock _ -> true | _ -> false) ->
+				| NormalAccess when 
+						(match follow f.cf_type with
+						| TFun _
+						| TDynamic _ ->
+							true;
+						| _ ->
+							false) -> 
+					newline ctx;
+					print ctx "%s::$%s = " (s_path ctx path false p) (s_ident f.cf_name);
+					gen_value ctx e
+				| MethodAccess true ->
 					newline ctx;
 					print ctx "%s::$%s = " (s_path ctx path false p) (s_ident f.cf_name);
 					gen_value ctx e
@@ -1633,13 +1778,6 @@ let generate_static_field_assign ctx path f =
 				print ctx "%s::$%s = " (s_path ctx path false p) (s_ident f.cf_name);
 				gen_value ctx e)
 
-let define_getset ctx stat f =
-	let def name =
-		Hashtbl.add ctx.get_sets (name,stat) f.cf_name
-	in
-		(match f.cf_get with MethodAccess m -> def m | _ -> ());
-		(match f.cf_set with MethodAccess m -> def m | _ -> ())
-
 let rec super_has_dynamic c =
 	match c.cl_super with
 	| None -> false
@@ -1650,8 +1788,7 @@ let rec super_has_dynamic c =
 let generate_class ctx c =
 	let requires_constructor = ref true in
 	ctx.curclass <- c;
-	List.iter (define_getset ctx false) c.cl_ordered_fields;
-	List.iter (define_getset ctx true) c.cl_ordered_statics;
+(*	ctx.curmethod <- ("new",true); *)
 	ctx.local_types <- List.map snd c.cl_types;
 
 	print ctx "%s %s " (if c.cl_interface then "interface" else "class") (s_path ctx c.cl_path c.cl_extern c.cl_pos);
@@ -1680,7 +1817,7 @@ let generate_class ctx c =
 			spr ctx "public function __construct(){}"
 		end;
 	| Some f ->
-	let f = { f with
+		let f = { f with
 			cf_name = "__construct";
 			cf_public = true;
 		} in
@@ -1693,12 +1830,12 @@ let generate_class ctx c =
 	(match c.cl_dynamic with
 		| Some _ when not c.cl_interface && not (super_has_dynamic c) ->
 			newline ctx;
-			spr ctx "private $__dynamics = array();\n\tpublic function &__get($n) {\n\t\tif(isset($this->__dynamics[$n]))\n\t\t\treturn $this->__dynamics[$n];\n\t}\n\tpublic function __set($n, $v) {\n\t\t$this->__dynamics[$n] = $v;\n\t}\n\tpublic function __call($n, $a) {\n\t\tif(is_callable($this->__dynamics[$n]))\n\t\t\treturn call_user_func_array($this->__dynamics[$n], $a);\n\t\tthrow new HException(\"Unable to call \".$n.\"\");\n\t}"
+			spr ctx "private $dynamics = array();\n\tpublic function &__get($n) {\n\t\tif(isset($this->dynamics[$n]))\n\t\t\treturn $this->dynamics[$n];\n\t}\n\tpublic function __set($n, $v) {\n\t\t$this->dynamics[$n] = $v;\n\t}\n\tpublic function __call($n, $a) {\n\t\tif(is_callable($this->dynamics[$n]))\n\t\t\treturn call_user_func_array($this->dynamics[$n], $a);\n\t\tthrow new HException(\"Unable to call \".$n.\"\");\n\t}"
 		| Some _
 		| _ ->
 			if List.length ctx.dynamic_methods > 0 then begin
 				newline ctx;
-				spr ctx "public function __call($m, $a) {\n\t\tif(isset($this->$m) && is_callable($this->$m))\n\t\t\treturn call_user_func_array($this->$m, $a);\n\t\telse if(isset($this->__dynamics[$m]) && is_callable($this->__dynamics[$m]))\n\t\t\treturn call_user_func_array($this->__dynamics[$m], $a);\n\t\telse\n\t\t\tthrow new HException('Unable to call '.$m.'');\n\t}";
+				spr ctx "public function __call($m, $a) {\n\t\tif(isset($this->$m) && is_callable($this->$m))\n\t\t\treturn call_user_func_array($this->$m, $a);\n\t\telse if(isset($this->dynamics[$m]) && is_callable($this->dynamics[$m]))\n\t\t\treturn call_user_func_array($this->dynamics[$m], $a);\n\t\telse\n\t\t\tthrow new HException('Unable to call '.$m.'');\n\t}";
 			end;
 	);
 
@@ -1718,11 +1855,12 @@ let generate_class ctx c =
 	end;
 	
 	print ctx "}"
-
+	
 let createmain com c =
 	let filename = match com.php_front with None -> "index.php" | Some n -> n in
 	let ctx = {
 		com = com;
+		stack = stack_init com false;
 		tabs = "";
 		ch = open_out (com.file ^ "/" ^ filename);
 		path = ([], "");
@@ -1734,11 +1872,11 @@ let createmain com c =
 		extern_required_paths = [];
 		extern_classes_with_init = [];
 		curclass = null_class;
+		curmethod = "";
 		locals = PMap.empty;
 		inv_locals = PMap.empty;
 		local_types = [];
 		inits = [];
-		get_sets = Hashtbl.create 0;
 		constructor_block = false;
 		quotes = 0;
 		dynamic_methods = [];
@@ -1863,6 +2001,12 @@ let generate com =
 					gen_expr ctx e);
 				List.iter (generate_static_field_assign ctx c.cl_path) c.cl_ordered_statics;
 				newline ctx;
+				if c.cl_path = (["php"], "Boot") & com.debug then begin
+					print ctx "$%s = new _hx_array(array())" ctx.stack.Codegen.stack_var;
+					newline ctx;
+					print ctx "$%s = new _hx_array(array())" ctx.stack.Codegen.stack_exc_var;
+					newline ctx;
+				end;
 				close ctx);
 		| TEnumDecl e ->
 			if e.e_extern then
@@ -1876,4 +2020,4 @@ let generate com =
 	) com.types;
 	Hashtbl.iter (fun name data ->
 		write_resource com.file name data
-	) com.resources;
\ No newline at end of file
+	) com.resources;
