@@ -296,23 +296,33 @@ let build_dependencies t =
 	let add_path p k =
 		h := PMap.add (p,k) () !h;
 	in
-	let rec add_type t =
-		match follow t with
+	let rec add_type_rec l t =
+		if List.memq t l then () else		
+		match t with
 		| TEnum (e,pl) ->
 			add_path e.e_path DKType;
-			List.iter add_type pl;
+			List.iter (add_type_rec (t::l)) pl;
 		| TInst (c,pl) ->
 			add_path c.cl_path DKType;
-			List.iter add_type pl;
-		| TFun (pl,t) ->
-			List.iter (fun (_,_,t) -> add_type t) pl;
-			add_type t;
+			List.iter (add_type_rec (t::l)) pl;
+		| TFun (pl,t2) ->
+			List.iter (fun (_,_,t2) -> add_type_rec (t::l) t2) pl;
+			add_type_rec (t::l) t2;
 		| TAnon a ->
-			PMap.iter (fun _ f -> add_type f.cf_type) a.a_fields
+			PMap.iter (fun _ f -> add_type_rec (t::l) f.cf_type) a.a_fields
 		| TDynamic t2 ->
-			if t2 != t then add_type t2;
-		| _ ->
-			()
+			add_type_rec (t::l) t2;
+		| TLazy f ->
+			add_type_rec l ((!f)())
+		| TMono r ->
+			(match !r with
+			| None -> ()
+			| Some t -> add_type_rec l t)
+		| TType (tt,pl) ->
+			add_type_rec (t::l) tt.t_type;
+			List.iter (add_type_rec (t::l)) pl
+	and add_type t = 
+		add_type_rec [] t
 	and add_expr e =
 		match e.eexpr with
 		| TTypeExpr t -> add_path (Type.t_path t) DKExpr
