@@ -589,13 +589,15 @@ let rec gen_access ?(read_write=false) ctx forcall e =
 			write ctx ASwap;
 			push ctx [p];
 		end;
-		(match follow e.etype with
-		| TFun _ -> VarClosure
-		| _ ->
-			if not !protect_all && Codegen.is_volatile e.etype then
-				VarVolatile
-			else
-				VarObj)
+		if not !protect_all && Codegen.is_volatile e.etype then
+			VarVolatile
+		else
+			VarObj
+	| TClosure (e,f) ->
+		gen_expr ctx true e;
+		if read_write then assert false;
+		push ctx [VStr (f,is_protected ctx e.etype f)];
+		VarClosure
 	| TArray (ea,eb) ->
 		if read_write then 
 			try 
@@ -888,7 +890,7 @@ and gen_call ctx e el =
 		write ctx (AGetURL2 (match post with [] -> 0 | [{ eexpr = TConst (TString "GET") }] -> 1 | _ -> 2))
 	| TLocal "__new__", e :: el ->
 		let nargs = List.length el in
-		List.iter (gen_expr ctx true) el;
+		List.iter (gen_expr ctx true) (List.rev el);
 		push ctx [VInt nargs];
 		let k = gen_access ctx true e in
 		new_call ctx k nargs
@@ -958,6 +960,7 @@ and gen_expr_2 ctx retval e =
 	| TConst TSuper
 	| TConst TThis
 	| TField _
+	| TClosure _
 	| TArray _
 	| TLocal _
 	| TTypeExpr _
