@@ -189,6 +189,7 @@ let rec follow_basic t =
 		| TInst ({ cl_path = (["haxe"],"Int32") },[])
 		| TInst ({ cl_path = ([],"Int") },[])
 		| TInst ({ cl_path = ([],"Float") },[])
+		| TType ({ t_path = [],"UInt" },[])
 		| TEnum ({ e_path = ([],"Bool") },[]) -> t
 		| t -> t)
 	| TType ({ t_path = [],"UInt" },[]) ->
@@ -418,16 +419,16 @@ let get_local_register ctx name =
 	| LReg r -> Some r
 	| _ -> None
 
-let rec setvar ctx (acc : write access) retval =
+let rec setvar ctx (acc : write access) kret =
 	match acc with
 	| VReg r ->
-		if retval then write ctx HDup;
+		if kret <> None then write ctx HDup;
 		set_reg ctx r;
-	| VGlobal _ | VId _ | VCast _ | VArray | VScope _ when retval ->
-		let r = alloc_reg ctx KDynamic in
+	| VGlobal _ | VId _ | VCast _ | VArray | VScope _ when kret <> None ->
+		let r = alloc_reg ctx (match kret with None -> assert false | Some k -> k) in
 		write ctx HDup;
 		set_reg ctx r;
-		setvar ctx acc false;
+		setvar ctx acc None;
 		write ctx (HReg r.rid);
 		free_reg ctx r
 	| VGlobal g ->
@@ -596,7 +597,7 @@ let begin_fun ctx args tret el stat p =
 		| acc ->
 			let r = alloc_reg ctx (classify ctx t) in
 			write ctx (HReg r.rid);
-			setvar ctx acc false
+			setvar ctx acc None
 	) args;
 	let args, varargs = (match args with
 		| ["__arguments__",_,_] -> [], true
@@ -733,7 +734,7 @@ let gen_access ctx e (forset : 'a) : 'a access =
 	match e.eexpr with
 	| TLocal i ->
 		gen_local_access ctx i e.epos forset
-	| TField (e1,f) ->
+	| TField (e1,f) | TClosure (e1,f) ->
 		let id, k, closure = property ctx f e1.etype in
 		if closure && not ctx.for_call then error "In Flash9, this method cannot be accessed this way : please define a local function" e1.epos;
 		(match e1.eexpr with
@@ -843,7 +844,7 @@ let rec gen_expr_content ctx retval e =
 		let id = type_path ctx (["flash"],"Error") in
 		write ctx (HFindPropStrict id);
 		write ctx (HConstructProperty (id,0));
-		setvar ctx (VId (ident "lastError")) false;
+		setvar ctx (VId (ident "lastError")) None;
 		gen_expr ctx true e;
 		write ctx HThrow;
 		no_value ctx retval;
@@ -885,7 +886,7 @@ let rec gen_expr_content ctx retval e =
 			| Some e ->
 				let acc = gen_local_access ctx v e.epos Write in
 				gen_expr ctx true e;
-				setvar ctx acc false)
+				setvar ctx acc None)
 		) vl
 	| TReturn None ->
 		write ctx HRetVoid;
@@ -897,6 +898,7 @@ let rec gen_expr_content ctx retval e =
 		ctx.infos.icond <- true;
 		no_value ctx retval
 	| TField _
+	| TClosure _
 	| TLocal _
 	| TTypeExpr _ ->
 		getvar ctx (gen_access ctx e Read)
@@ -990,7 +992,7 @@ let rec gen_expr_content ctx retval e =
 				) in
 				let acc = gen_local_access ctx ename e.epos Write in
 				(match r with None -> () | Some r -> write ctx (HReg r.rid));
-				setvar ctx acc false;
+				setvar ctx acc None;
 				(* ----- *)
 				let rec call_loop e =
 					match e.eexpr with
@@ -1005,7 +1007,7 @@ let rec gen_expr_content ctx retval e =
 					let j = jump ctx J3False in
 					getvar ctx (VGlobal (type_path ctx (["flash"],"Boot")));
 					getvar ctx (gen_local_access ctx ename e.epos Read);
-					setvar ctx (VId (ident "lastError")) false;
+					setvar ctx (VId (ident "lastError")) None;
 					j();
 				end;
 				gen_expr ctx retval e;
@@ -1037,7 +1039,7 @@ let rec gen_expr_content ctx retval e =
 		let acc = gen_local_access ctx v e.epos Write in
 		write ctx (HReg r.rid);
 		write ctx (HCallProperty (ident "next",0));
-		setvar ctx acc false;
+		setvar ctx acc None;
 		gen_expr ctx false e;
 		start J3Always;
 		end_loop continue_pos;
@@ -1181,7 +1183,7 @@ let rec gen_expr_content ctx retval e =
 						write ctx (HReg rparams.rid);
 						write ctx (HSmallInt !p);
 						getvar ctx VArray;
-						setvar ctx acc false
+						setvar ctx acc None
 				) l
 			);
 			gen_expr ctx retval e;
@@ -1396,12 +1398,12 @@ and gen_unop ctx retval op flag e =
 			write ctx HDup;
 			set_reg ctx r;
 			write ctx (HOp op);
-			setvar ctx acc_write false;
+			setvar ctx acc_write None;
 			write ctx (HReg r.rid);
 			free_reg ctx r
 		| Postfix | Prefix ->
 			write ctx (HOp op);
-			setvar ctx acc_write retval
+			setvar ctx acc_write (if retval then Some k else None)
 
 and gen_binop ctx retval op e1 e2 t =
 	let write_op op =		
@@ -1454,7 +1456,7 @@ and gen_binop ctx retval op e1 e2 t =
 	| OpAssign ->
 		let acc = gen_access ctx e1 Write in
 		gen_expr ctx true e2;
-		setvar ctx acc retval
+		setvar ctx acc (if retval then Some (classify ctx e1.etype) else None)
 	| OpBoolAnd ->
 		write ctx HFalse;
 		let j = jump_expr ctx e1 false in
@@ -1478,7 +1480,7 @@ and gen_binop ctx retval op e1 e2 t =
 		getvar ctx racc;
 		gen_expr ctx true e2;
 		write_op op;
-		setvar ctx wacc retval
+		setvar ctx wacc (if retval then Some (classify ctx e1.etype) else None)
 	| OpAdd | OpMult | OpDiv | OpSub | OpAnd | OpOr | OpXor | OpShl | OpShr | OpUShr | OpMod ->
 		gen_expr ctx true e1;
 		gen_expr ctx true e2;
@@ -1586,7 +1588,7 @@ let generate_construct ctx fdata c =
 	(* --- *)
 	PMap.iter (fun _ f ->
 		match f.cf_expr with
-		| Some { eexpr = TFunction fdata } when f.cf_set = NormalAccess ->
+		| Some { eexpr = TFunction fdata } when f.cf_set = MethodAccess true ->
 			let id = ident f.cf_name in
 			write ctx (HFindProp id);
 			write ctx (HFunction (generate_method ctx fdata false));
@@ -1622,7 +1624,7 @@ let generate_class_init ctx c hc =
 	write ctx (HClassDef hc);
 	List.iter (fun f ->
 		match f.cf_expr with
-		| Some { eexpr = TFunction fdata } when f.cf_set = NormalAccess ->
+		| Some { eexpr = TFunction fdata } when f.cf_set = MethodAccess true ->
 			write ctx HDup;
 			write ctx (HFunction (generate_method ctx fdata true));
 			write ctx (HInitProp (ident f.cf_name));
@@ -1674,7 +1676,7 @@ let generate_field_kind ctx f c stat =
 			| Some (c,_) ->
 				PMap.exists f.cf_name c.cl_fields || loop c
 		in
-		if f.cf_set = NormalAccess then
+		if f.cf_set = NormalAccess || f.cf_set = MethodAccess true then
 			Some (HFVar {
 				hlv_type = Some (type_path ctx ([],"Function"));
 				hlv_value = HVNone;
@@ -1689,7 +1691,7 @@ let generate_field_kind ctx f c stat =
 			})
 	| _ when c.cl_interface && not stat ->
 		(match follow f.cf_type with
-		| TFun (args,tret) when f.cf_set = MethodCantAccess ->
+		| TFun (args,tret) when f.cf_set = MethodAccess false ->
 			Some (HFMethod {
 				hlm_type = end_fun ctx (List.map (fun (a,opt,t) -> a, (if opt then Some TNull else None), t) args) tret;
 				hlm_final = false;
