@@ -23,11 +23,11 @@ type path = string list * string
 type field_access =
 	| NormalAccess
 	| NoAccess
-	| ResolveAccess
-	| MethodAccess of string
-	| MethodCantAccess
-	| NeverAccess
-	| InlineAccess
+	| ResolveAccess (* call resolve("field") when accessed *)
+	| CallAccess of string (* perform a method call when accessed *)
+	| MethodAccess of bool (* true = the method is dynamic *)
+	| InlineAccess (* similar to Normal but inline when acccessed *)
+	| NeverAccess (* can't be accessed, even in subclasses *)
 
 type t =
 	| TMono of t option ref
@@ -75,6 +75,7 @@ and texpr_expr =
 	| TArray of texpr * texpr
 	| TBinop of Ast.binop * texpr * texpr
 	| TField of texpr * string
+	| TClosure of texpr * string
 	| TTypeExpr of module_type
 	| TParenthesis of texpr
 	| TObjectDecl of (string * texpr) list
@@ -289,8 +290,8 @@ let s_access = function
 	| NormalAccess -> "default"
 	| NoAccess -> "null"
 	| NeverAccess -> "never"
-	| MethodAccess m -> m
-	| MethodCantAccess -> "dynamic"
+	| CallAccess m -> m
+	| MethodAccess b -> if b then "dynamic method" else "default method"
 	| ResolveAccess -> "resolve"
 	| InlineAccess -> "inline"
 
@@ -490,16 +491,24 @@ let has_no_field t n = Has_no_field (t,n)
 let has_extra_field t n = Has_extra_field (t,n)
 let error l = raise (Unify_error l)
 
+type simple_access =
+	| SAYes
+	| SANo
+	| SARuntime
+
+let simple_access = function	
+	| NormalAccess | InlineAccess | MethodAccess true -> SAYes
+	| NoAccess | NeverAccess | MethodAccess false -> SANo
+	| ResolveAccess | CallAccess _ -> SARuntime
+
+(*
+	we can restrict access as soon as both are runtime-compatible
+*)
 let unify_access a1 a2 =
-	a1 = a2 || match a1, a2 with
-	| NormalAccess, NoAccess
-	| NormalAccess, MethodCantAccess
-	| NormalAccess, NeverAccess
-	| MethodCantAccess, NoAccess
-	| NeverAccess, NoAccess
-	| MethodCantAccess, NeverAccess
-	| NoAccess, NeverAccess -> true
-	| _ -> false
+	a1 = a2 || match simple_access a1 , simple_access a2 with
+		| SAYes, SAYes
+		| _, SANo -> true
+		| _ -> false
 
 let eq_stack = ref []
 
@@ -731,9 +740,9 @@ let rec unify a b =
 		(match !(an.a_status) with
 		| Statics cl -> unify (TInst (cl,List.map snd cl.cl_types)) pt
 		| _ -> error [cannot_unify a b])
-	| TAnon an, TInst ({ cl_path = [],"Enum" },[]) ->
+	| TAnon an, TInst ({ cl_path = [],"Enum" },[pt]) ->
 		(match !(an.a_status) with
-		| EnumStatics _ -> ()
+		| EnumStatics e -> unify (TEnum (e,List.map snd e.e_types)) pt
 		| _ -> error [cannot_unify a b])
 	| TDynamic t , _ ->
 		if t == a then
@@ -791,6 +800,7 @@ let iter f e =
 		f e2;
 	| TThrow e
 	| TField (e,_)
+	| TClosure (e,_)
 	| TParenthesis e
 	| TUnop (_,_,e) ->
 		f e
@@ -846,6 +856,8 @@ let map_expr f e =
 		{ e with eexpr = TThrow (f e1) }
 	| TField (e1,v) ->
 		{ e with eexpr = TField (f e1,v) }
+	| TClosure (e1,v) ->
+		{ e with eexpr = TClosure (f e1,v) }
 	| TParenthesis e1 ->
 		{ e with eexpr = TParenthesis (f e1) }
 	| TUnop (op,pre,e1) ->
@@ -896,6 +908,8 @@ let map_expr_type f ft e =
 		{ e with eexpr = TThrow (f e1); etype = ft e.etype }
 	| TField (e1,v) ->
 		{ e with eexpr = TField (f e1,v); etype = ft e.etype }
+	| TClosure (e1,v) ->
+		{ e with eexpr = TClosure (f e1,v); etype = ft e.etype }
 	| TParenthesis e1 ->
 		{ e with eexpr = TParenthesis (f e1); etype = ft e.etype }
 	| TUnop (op,pre,e1) ->
