@@ -24,6 +24,7 @@ open Typecore
 let type_constant ctx c p =
 	match c with
 	| Int s ->
+		if String.length s > 10 && String.sub s 0 2 = "0x" then error "Invalid hexadecimal integer" p;
 		(try
 			mk (TConst (TInt (Int32.of_string s))) ctx.api.tint p
 		with
@@ -195,15 +196,15 @@ and load_type ctx p t =
 				| AFFun (tl,t) ->
 					let t = load_type ctx p t in
 					let args = List.map (fun (name,o,t) -> name , o, load_type ctx p t) tl in
-					TFun (args,t), NormalAccess, MethodCantAccess
+					TFun (args,t), NormalAccess, MethodAccess false
 				| AFProp (t,i1,i2) ->
 					let access m get =
 						match m with
 						| "null" -> NoAccess
 						| "never" -> NeverAccess
 						| "default" -> NormalAccess
-						| "dynamic" -> MethodAccess ((if get then "get_"  else "set_") ^ n)
-						| _ -> MethodAccess m
+						| "dynamic" -> CallAccess ((if get then "get_"  else "set_") ^ n)
+						| _ -> CallAccess m
 					in
 					load_type ctx p t, access i1 true, access i2 false
 			) in
@@ -337,8 +338,10 @@ let rec check_interface ctx c p intf params =
 			let p = (match f2.cf_expr with None -> p | Some e -> e.epos) in
 			if f.cf_public && not f2.cf_public then
 				display_error ctx ("Field " ^ i ^ " should be public as requested by " ^ s_type_path intf.cl_path) p
-			else if f2.cf_get <> f.cf_get || (f2.cf_set <> f.cf_set && (f2.cf_set,f.cf_set) <> (NeverAccess,NoAccess)) then
-				display_error ctx ("Field " ^ i ^ " has different property access than in " ^ s_type_path intf.cl_path) p
+			else if not (match f2.cf_get, f.cf_get with InlineAccess, NormalAccess -> true | a,b -> a = b) then
+				display_error ctx ("Field " ^ i ^ " has different property access than in " ^ s_type_path intf.cl_path ^ " (" ^ s_access f2.cf_get ^ " should be " ^ s_access f.cf_get ^ ")") p
+			else if not (match f2.cf_set, f.cf_set with NeverAccess, NoAccess -> true | NeverAccess, MethodAccess false -> f2.cf_get = InlineAccess | a,b -> a = b) then
+				display_error ctx ("Field " ^ i ^ " has different property access than in " ^ s_type_path intf.cl_path ^ " (" ^ s_access f2.cf_set ^ " should be " ^ s_access f.cf_set ^ ")") p			
 			else try
 				valid_redefinition ctx f2 t2 f (apply_params intf.cl_types params f.cf_type)
 			with
@@ -615,8 +618,9 @@ let init_class ctx c p herits fields =
 			) params in
 			let stat = List.mem AStatic access in
 			let inline = List.mem AInline access in
+			if inline && c.cl_interface then error "You can't declare inline methods in interfaces" p;
 			let parent = (if not stat then get_parent c name else None) in
-			let dynamic = List.mem ADynamic access || (match parent with Some { cf_set = NormalAccess } -> true | _ -> false) in
+			let dynamic = List.mem ADynamic access || (match parent with Some { cf_set = MethodAccess true } -> true | _ -> false) in
 			let ctx = { ctx with
 				curclass = c;
 				curmethod = name;
@@ -641,7 +645,7 @@ let init_class ctx c p herits fields =
 				cf_doc = doc;
 				cf_type = t;
 				cf_get = if inline then InlineAccess else NormalAccess;
-				cf_set = (if inline then NeverAccess else if dynamic then NormalAccess else MethodCantAccess);
+				cf_set = (if inline then NeverAccess else MethodAccess dynamic);
 				cf_expr = None;
 				cf_public = is_public access parent;
 				cf_params = params;
@@ -682,12 +686,12 @@ let init_class ctx c p herits fields =
 			in
 			let get = (match get with
 				| "null" -> NoAccess
-				| "dynamic" -> MethodAccess ("get_" ^ name)
+				| "dynamic" -> CallAccess ("get_" ^ name)
 				| "never" -> NeverAccess
 				| "default" -> NormalAccess
 				| _ ->
 					check_get := check_method get (TFun ([],ret));
-					MethodAccess get
+					CallAccess get
 			) in
 			let set = (match set with
 				| "null" ->
@@ -697,13 +701,13 @@ let init_class ctx c p herits fields =
 					else
 						NoAccess
 				| "never" -> NeverAccess
-				| "dynamic" -> MethodAccess ("set_" ^ name)
+				| "dynamic" -> CallAccess ("set_" ^ name)
 				| "default" -> NormalAccess
 				| _ ->
 					check_set := check_method set (TFun (["",false,ret],ret));
-					MethodAccess set
+					CallAccess set
 			) in
-			if set = NormalAccess && (match get with MethodAccess _ -> true | _ -> false) then error "Unsupported property combination" p;
+			if set = NormalAccess && (match get with CallAccess _ -> true | _ -> false) then error "Unsupported property combination" p;
 			let cf = {
 				cf_name = name;
 				cf_doc = doc;
@@ -726,6 +730,7 @@ let init_class ctx c p herits fields =
 			c.cl_constructor <- Some f;
 		end else if not is_static || f.cf_name <> "__init__" then begin
 			if PMap.mem f.cf_name (if is_static then c.cl_statics else c.cl_fields) then error ("Duplicate class field declaration : " ^ f.cf_name) p;
+			if PMap.exists f.cf_name (if is_static then c.cl_fields else c.cl_statics) then error ("Same field name can't be use for both static and instance : " ^ f.cf_name) p;
 			if is_static then begin
 				c.cl_statics <- PMap.add f.cf_name f c.cl_statics;
 				c.cl_ordered_statics <- f :: c.cl_ordered_statics;
@@ -822,7 +827,7 @@ let type_module ctx m tdecls loadp =
 	in
 	List.iter (fun (d,p) ->
 		match d with
-		| EImport _ -> ()
+		| EImport _ | EUsing _ -> ()
 		| EClass d ->
 			let priv = List.mem HPrivate d.d_flags in
 			let path = decl_with_name d.d_name p priv in
@@ -885,6 +890,7 @@ let type_module ctx m tdecls loadp =
 		locals_map = PMap.empty;
 		locals_map_inv = PMap.empty;
 		local_types = ctx.std.mtypes @ m.mtypes;
+		local_using = [];
 		type_params = [];
 		curmethod = "";
 		super_call = false;
@@ -912,7 +918,7 @@ let type_module ctx m tdecls loadp =
 	(* here is an additional PASS 1 phase, which handle the type parameters declaration, with lazy contraints *)
 	List.iter (fun (d,p) ->
 		match d with
-		| EImport _ -> ()
+		| EImport _ | EUsing _ -> ()
 		| EClass d ->
 			let c = get_class d.d_name in
 			c.cl_types <- List.map (type_type_params ctx c.cl_path p) d.d_params;
@@ -939,6 +945,10 @@ let type_module ctx m tdecls loadp =
 					Not_found -> error ("Module " ^ s_type_path (pack,name) ^ " does not define type " ^ t) p
 			);
 			m.mimports <- (md,topt) :: m.mimports;
+		| EUsing (pack,name) ->
+			let md = ctx.api.load_module (pack,name) p in
+			let types = List.filter (fun t -> not (t_private t)) md.mtypes in
+			ctx.local_using <- ctx.local_using @ types;
 		| EClass d ->
 			let c = get_class d.d_name in
 			delays := !delays @ check_overriding ctx c p :: check_interfaces ctx c p :: init_class ctx c p d.d_flags d.d_data
@@ -952,7 +962,13 @@ let type_module ctx m tdecls loadp =
 				if c = "name" && Common.defined ctx.com "js" then error "This identifier cannot be used in Javascript" p;
 				let t = (match t with
 					| [] -> et
-					| l -> TFun (List.map (fun (s,opt,t) -> s, opt, load_type_opt ~opt ctx p (Some t)) l, et)
+					| l ->
+						let pnames = ref PMap.empty in
+						TFun (List.map (fun (s,opt,t) -> 
+							if PMap.mem s (!pnames) then error ("Duplicate parameter '" ^ s ^ "' in enum constructor " ^ c) p;
+							pnames := PMap.add s () (!pnames);
+							s, opt, load_type_opt ~opt ctx p (Some t)
+						) l, et)
 				) in
 				if PMap.mem c e.e_constrs then error ("Duplicate constructor " ^ c) p;
 				e.e_constrs <- PMap.add c {
@@ -1035,7 +1051,7 @@ let parse_module ctx m p =
 			| EClass d -> build HPrivate d
 			| EEnum d -> build EPrivate d
 			| ETypedef d -> build EPrivate d
-			| EImport _ -> acc
+			| EImport _ | EUsing _ -> acc
 		) [(EImport (!remap, snd m, None),null_pos)] decls)
 	else
 		decls
