@@ -28,6 +28,11 @@ type switch_mode =
 	| CMatch of (tenum_field * (string option * t) list option)
 	| CExpr of texpr
 
+type access_mode =
+	| MGet
+	| MSet
+	| MCall
+
 exception Display of t
 
 type access_kind =
@@ -35,6 +40,7 @@ type access_kind =
 	| AccExpr of texpr
 	| AccSet of texpr * string * t * string
 	| AccInline of texpr * tclass_field * t
+	| AccUsing of texpr * texpr
 
 let mk_infos ctx p params =
 	(EObjectDecl (
@@ -89,6 +95,8 @@ let classify t =
 	| TDynamic _ -> KDyn
 	| _ -> KOther
 
+let type_field_rec = ref (fun _ _ _ _ _ -> assert false)
+
 (* ---------------------------------------------------------------------- *)
 (* PASS 3 : type expression & check structure *)
 
@@ -114,7 +122,7 @@ let unify_call_params ctx name el args p inline =
 		let argstr = "Function " ^ (match name with None -> "" | Some n -> "'" ^ n ^ "' ") ^ "requires " ^ (if args = [] then "no arguments" else "arguments : " ^ String.concat ", " (List.map format_arg args)) in
 		display_error ctx (txt ^ " arguments\n" ^ argstr) p
 	in
-	let arg_error ul name opt =
+	let arg_error ul name opt p =
 		raise (Error (Stack (Unify ul,Custom ("For " ^ (if opt then "optional " else "") ^ "function argument '" ^ name ^ "'")), p))
 	in
 	let rec no_opt = function
@@ -159,7 +167,7 @@ let unify_call_params ctx name el args p inline =
 		| _ , [] ->
 			(match List.rev skip with
 			| [] -> error "Too many"
-			| [name,ul] -> arg_error ul name true
+			| [name,ul] -> arg_error ul name true p
 			| _ -> error "Invalid");
 			[]
 		| ee :: l, (name,opt,t) :: l2 ->
@@ -171,8 +179,8 @@ let unify_call_params ctx name el args p inline =
 				Error (Unify ul,_) ->
 					if opt then
 						loop (default_value t :: acc) (ee :: l) l2 ((name,ul) :: skip)
-					else
-						arg_error ul name false
+					else						
+						arg_error ul name false e.epos
 	in
 	loop [] el args []
 
@@ -182,8 +190,7 @@ let type_local ctx i p =
 	let i = (try PMap.find i ctx.locals_map with Not_found -> i) in
 	mk (TLocal i) t p
 
-let type_type ctx tpath p =
-	let rec loop t tparams =
+let rec type_module_type ctx t tparams p =
 	match t with
 	| TClassDecl c ->
 		let t_tmp = {
@@ -197,7 +204,9 @@ let type_type ctx tpath p =
 			t_private = true;
 			t_types = [];
 		} in
-		mk (TTypeExpr (TClassDecl c)) (TType (t_tmp,[])) p
+		let e = mk (TTypeExpr (TClassDecl c)) (TType (t_tmp,[])) p in
+		check_locals_masking ctx e;
+		e
 	| TEnumDecl e ->
 		let types = (match tparams with None -> List.map (fun _ -> mk_mono()) e.e_types | Some l -> l) in
 		let fl = PMap.fold (fun f acc ->
@@ -206,7 +215,7 @@ let type_type ctx tpath p =
 				cf_public = true;
 				cf_type = f.ef_type;
 				cf_get = NormalAccess;
-				cf_set = NoAccess;
+				cf_set = (match follow f.ef_type with TFun _ -> MethodAccess false | _ -> NoAccess);
 				cf_doc = None;
 				cf_expr = None;
 				cf_params = [];
@@ -223,20 +232,21 @@ let type_type ctx tpath p =
 			t_private = true;
 			t_types = e.e_types;
 		} in
-		mk (TTypeExpr (TEnumDecl e)) (TType (t_tmp,types)) p
+		let e = mk (TTypeExpr (TEnumDecl e)) (TType (t_tmp,types)) p in
+		check_locals_masking ctx e;
+		e
 	| TTypeDecl s ->
 		let t = apply_params s.t_types (List.map (fun _ -> mk_mono()) s.t_types) s.t_type in
 		match follow t with
 		| TEnum (e,params) ->
-			loop (TEnumDecl e) (Some params)
+			type_module_type ctx (TEnumDecl e) (Some params) p
 		| TInst (c,params) ->
-			loop (TClassDecl c) (Some params)
+			type_module_type ctx (TClassDecl c) (Some params) p
 		| _ ->
-			error (s_type_path tpath ^ " is not a value") p
-	in
-	let e = loop (Typeload.load_type_def ctx p tpath) None in
-	check_locals_masking ctx e;
-	e
+			error (s_type_path s.t_path ^ " is not a value") p
+
+let type_type ctx tpath p =
+	type_module_type ctx (Typeload.load_type_def ctx p tpath) None p
 
 let get_constructor c p =
 	let rec loop c = 
@@ -254,22 +264,62 @@ let get_constructor c p =
 	with Not_found ->
 		error (s_type_path c.cl_path ^ " does not have a constructor") p
 
-let acc_get g p =
+let make_call ctx e params t p =
+	try
+		if not ctx.doinline then raise Exit;
+		let ethis, fname = (match e.eexpr with TField (ethis,fname) -> ethis, fname | _ -> raise Exit) in
+		let f = (match follow ethis.etype with
+			| TInst (c,params) -> snd (try class_field c fname with Not_found -> raise Exit)
+			| TAnon a -> (try PMap.find fname a.a_fields with Not_found -> raise Exit)
+			| _ -> raise Exit
+		) in
+		if f.cf_get <> InlineAccess then raise Exit;
+		ignore(follow f.cf_type); (* force evaluation *)
+		(match f.cf_expr with
+		| Some { eexpr = TFunction fd } ->
+			(match Optimizer.type_inline ctx f fd ethis params t p with
+			| None -> raise Exit
+			| Some e -> e)
+		| _ -> 
+			error "Recursive inline is not supported" p)
+	with Exit ->
+		mk (TCall (e,params)) t p
+		
+let rec acc_get ctx g p =
 	match g with
 	| AccNo f -> error ("Field " ^ f ^ " cannot be accessed for reading") p
 	| AccExpr e -> e
 	| AccSet _ -> assert false
+	| AccUsing (et,e) ->
+		(* build a closure with first parameter applied *)
+		(match follow et.etype with
+		| TFun (_ :: args,ret) ->
+			let tcallb = TFun (args,ret) in	
+			let twrap = TFun ([("_e",false,e.etype)],tcallb) in
+			let ecall = make_call ctx et (List.map (fun (n,_,t) -> mk (TLocal n) t p) (("_e",false,e.etype) :: args)) ret p in
+			let ecallb = mk (TFunction {
+				tf_args = List.map (fun (n,_,t) -> n,None,t) args;
+				tf_type = ret;
+				tf_expr = mk (TReturn (Some ecall)) t_dynamic p;
+			}) tcallb p in
+			let ewrap = mk (TFunction {
+				tf_args = [("_e",None,e.etype)];
+				tf_type = tcallb;
+				tf_expr = mk (TReturn (Some ecallb)) t_dynamic p; 
+			}) twrap p in
+			make_call ctx ewrap [e] tcallb p
+		| _ -> assert false)
 	| AccInline (e,f,t) ->
 		ignore(follow f.cf_type); (* force computing *)
 		match f.cf_expr with
 		| None -> error "Recursive inline is not supported" p
-		| Some { eexpr = TFunction _ } ->  mk (TField (e,f.cf_name)) t p
+		| Some { eexpr = TFunction _ } ->  mk (TClosure (e,f.cf_name)) t p
 		| Some e -> 
 			let rec loop e = Type.map_expr loop { e with epos = p } in
 			loop e
 
-let field_access ctx get f t e p =
-	match if get then f.cf_get else f.cf_set with
+let field_access ctx mode f t e p =
+	match (match mode with MGet | MCall -> f.cf_get | MSet -> f.cf_set) with
 	| NoAccess ->
 		let normal = AccExpr (mk (TField (e,f.cf_name)) t p) in
 		(match follow e.etype with
@@ -280,68 +330,73 @@ let field_access ctx get f t e p =
 			| _ -> if ctx.untyped then normal else AccNo f.cf_name)
 		| _ ->
 			if ctx.untyped then normal else AccNo f.cf_name)
-	| MethodCantAccess when not ctx.untyped ->
+	| MethodAccess false when not ctx.untyped ->
 		error "Cannot rebind this method : please use 'dynamic' before method declaration" p
-	| NormalAccess | MethodCantAccess ->
-		AccExpr (mk (TField (e,f.cf_name)) t p)
-	| MethodAccess m ->
+	| NormalAccess | MethodAccess _ ->
+		(* 
+			creates a closure if we're reading a normal method
+			or a read-only variable (which could be a method)
+		*)
+		(match mode, f.cf_set with
+		| MGet, MethodAccess _ -> AccExpr (mk (TClosure (e,f.cf_name)) t p)
+		| MGet, NoAccess | MGet, NeverAccess when (match follow t with TFun _ -> true | _ -> false) -> AccExpr (mk (TClosure (e,f.cf_name)) t p)
+		| _ -> AccExpr (mk (TField (e,f.cf_name)) t p))
+	| CallAccess m ->
 		if m = ctx.curmethod && (match e.eexpr with TConst TThis -> true | TTypeExpr (TClassDecl c) when c == ctx.curclass -> true | _ -> false) then
-			let prefix = if Common.defined ctx.com "as3gen" then "$" else "" in
+			let prefix = if Common.defined ctx.com "as3" then "$" else "" in
 			AccExpr (mk (TField (e,prefix ^ f.cf_name)) t p)
-		else if get then
-			AccExpr (mk (TCall (mk (TField (e,m)) (tfun [] t) p,[])) t p)
-		else
+		else if mode = MSet then
 			AccSet (e,m,t,f.cf_name)
+		else
+			AccExpr (make_call ctx (mk (TField (e,m)) (tfun [] t) p) [] t p)
 	| ResolveAccess ->
 		let fstring = mk (TConst (TString f.cf_name)) ctx.api.tstring p in
 		let tresolve = tfun [ctx.api.tstring] t in
-		AccExpr (mk (TCall (mk (TField (e,"resolve")) tresolve p,[fstring])) t p)
+		AccExpr (make_call ctx (mk (TField (e,"resolve")) tresolve p) [fstring] t p)
 	| NeverAccess ->
 		AccNo f.cf_name
 	| InlineAccess ->
 		AccInline (e,f,t)
 
-let type_ident ctx i is_type p get =
+let type_ident ctx i is_type p mode =
 	match i with
 	| "true" ->
-		if get then
+		if mode = MGet then
 			AccExpr (mk (TConst (TBool true)) ctx.api.tbool p)
 		else
 			AccNo i
 	| "false" ->
-		if get then
+		if mode = MGet then
 			AccExpr (mk (TConst (TBool false)) ctx.api.tbool p)
 		else
 			AccNo i
 	| "this" ->
 		if not ctx.untyped && ctx.in_static then error "Cannot access this from a static function" p;
-		if get then
+		if mode = MGet then
 			AccExpr (mk (TConst TThis) ctx.tthis p)
 		else
 			AccNo i
 	| "super" ->
-		if not ctx.super_call then
-			AccNo i
-		else
 		let t = (match ctx.curclass.cl_super with
-		| None -> error "Current class does not have a superclass" p
-		| Some (c,params) -> TInst(c,params)
+			| None -> error "Current class does not have a superclass" p
+			| Some (c,params) -> TInst(c,params)
 		) in
 		if ctx.in_static then error "Cannot access super from a static function" p;
-		ctx.super_call <- false;
-		if get then
-			AccExpr (mk (TConst TSuper) t p)
-		else
+		if mode = MSet || not ctx.super_call then
 			AccNo i
+		else begin
+			ctx.super_call <- false;
+			AccExpr (mk (TConst TSuper) t p)
+		end
 	| "null" ->
-		if get then
+		if mode = MGet then
 			AccExpr (null (mk_mono()) p)
 		else
 			AccNo i
 	| "here" ->
 		let infos = mk_infos ctx p [] in
 		let e = type_expr ctx infos true in
-		if get then
+		if mode = MGet then
 			AccExpr { e with etype = Typeload.load_normal_type ctx { tpackage = ["haxe"]; tname = "PosInfos"; tparams = [] } p false }
 		else
 			AccNo i
@@ -353,12 +408,12 @@ let type_ident ctx i is_type p get =
 		(* member variable lookup *)
 		if ctx.in_static then raise Not_found;
 		let t , f = class_field ctx.curclass i in
-		field_access ctx get f t (mk (TConst TThis) ctx.tthis p) p
+		field_access ctx mode f t (mk (TConst TThis) ctx.tthis p) p
 	with Not_found -> try
 		(* static variable lookup *)
 		let f = PMap.find i ctx.curclass.cl_statics in
 		let e = type_type ctx ctx.curclass.cl_path p in
-		field_access ctx get f (field_type f) e p
+		field_access ctx mode f (field_type f) e p
 	with Not_found -> try
 		(* lookup imported *)
 		let rec loop l =
@@ -377,10 +432,10 @@ let type_ident ctx i is_type p get =
 		in
 		let e = loop ctx.local_types in
 		check_locals_masking ctx e;
-		if get then
-			AccExpr e
-		else
+		if mode = MSet then
 			AccNo i
+		else
+			AccExpr e
 	with Not_found -> try
 		(* lookup type *)
 		if not is_type then raise Not_found;
@@ -437,7 +492,29 @@ let type_matching ctx (enum,params) (e,p) ecases first_case =
 	| _ ->
 		invalid()
 
-let type_field ctx e i p get =
+let rec type_field ctx e i p mode =
+	let using_field() =
+		if mode = MSet then raise Not_found;
+		let rec loop = function
+			| [] ->
+				raise Not_found
+			| TEnumDecl _ :: l | TTypeDecl _ :: l ->
+				loop l
+			| TClassDecl c :: l ->
+				try
+					let f = PMap.find i c.cl_statics in
+					let t = field_type f in
+					(match follow t with
+					| TFun ((_,_,t0) :: args,r) ->
+						(try unify_raise ctx e.etype t0 p with Error (Unify _,_) -> raise Not_found);
+						let et = type_module_type ctx (TClassDecl c) None p in						
+						AccUsing (mk (TField (et,i)) t p,e)
+					| _ -> raise Not_found)
+				with Not_found ->
+					loop l
+		in
+		loop ctx.local_using
+	in
 	let no_field() =
 		if not ctx.untyped then display_error ctx (s_type (print_context()) e.etype ^ " has no field " ^ i) p;
 		AccExpr (mk (TField (e,i)) (mk_mono()) p)
@@ -448,8 +525,8 @@ let type_field ctx e i p get =
 			match c.cl_dynamic with
 			| Some t ->
 				let t = apply_params c.cl_types params t in
-				if get && PMap.mem "resolve" c.cl_fields then
-					AccExpr (mk (TCall (mk (TField (e,"resolve")) (tfun [ctx.api.tstring] t) p,[Typeload.type_constant ctx (String i) p])) t p)
+				if mode = MGet && PMap.mem "resolve" c.cl_fields then
+					AccExpr (make_call ctx (mk (TField (e,"resolve")) (tfun [ctx.api.tstring] t) p) [Typeload.type_constant ctx (String i) p] t p)
 				else
 					AccExpr (mk (TField (e,i)) t p)
 			| None ->
@@ -461,7 +538,9 @@ let type_field ctx e i p get =
 			let t , f = class_field c i in
 			if e.eexpr = TConst TSuper && f.cf_set = NormalAccess && Common.platform ctx.com Flash9 then error "Cannot access superclass variable for calling : needs to be a proper method" p;
 			if not f.cf_public && not (is_parent c ctx.curclass) && not ctx.untyped then display_error ctx ("Cannot access to private field " ^ i) p;
-			field_access ctx get f (apply_params c.cl_types params t) e p
+			field_access ctx mode f (apply_params c.cl_types params t) e p
+		with Not_found -> try
+			using_field()
 		with Not_found -> try
 			loop_dyn c params
 		with Not_found ->
@@ -478,9 +557,11 @@ let type_field ctx e i p get =
 				| Statics c when is_parent c ctx.curclass -> ()
 				| _ -> display_error ctx ("Cannot access to private field " ^ i) p
 			end;
-			field_access ctx get f (field_type f) e p
+			field_access ctx mode f (field_type f) e p
 		with Not_found ->
-			if is_closed a then
+			if is_closed a then try
+				using_field()
+			with Not_found ->
 				no_field()
 			else
 			let f = {
@@ -489,12 +570,12 @@ let type_field ctx e i p get =
 				cf_doc = None;
 				cf_public = true;
 				cf_get = NormalAccess;
-				cf_set = if get then NoAccess else NormalAccess;
+				cf_set = (match mode with MSet -> NormalAccess | MGet | MCall -> NoAccess);
 				cf_expr = None;
 				cf_params = [];
 			} in
 			a.a_fields <- PMap.add i f a.a_fields;
-			field_access ctx get f (field_type f) e p
+			field_access ctx mode f (field_type f) e p
 		)
 	| TMono r ->
 		if ctx.untyped && Common.defined ctx.com "swf-mark" && Common.defined ctx.com "flash" then ctx.com.warning "Mark" p;
@@ -504,7 +585,7 @@ let type_field ctx e i p get =
 			cf_doc = None;
 			cf_public = true;
 			cf_get = NormalAccess;
-			cf_set = if get then NoAccess else NormalAccess;
+			cf_set = (match mode with MSet -> NormalAccess | MGet | MCall -> NoAccess);
 			cf_expr = None;
 			cf_params = [];
 		} in
@@ -512,9 +593,9 @@ let type_field ctx e i p get =
 		let t = TAnon { a_fields = PMap.add i f PMap.empty; a_status = x } in
 		ctx.opened <- x :: ctx.opened;
 		r := Some t;
-		field_access ctx get f (field_type f) e p
+		field_access ctx mode f (field_type f) e p
 	| t ->
-		no_field()
+		try using_field() with Not_found -> no_field()
 
 (*
 	We want to try unifying as an integer and apply side effects.
@@ -563,8 +644,8 @@ let unify_int ctx e k =
 let rec type_binop ctx op e1 e2 p =
 	match op with
 	| OpAssign ->
-		let e1 = type_access ctx (fst e1) (snd e1) false in
-		let e2 = type_expr_with_type ctx e2 (match e1 with AccNo _ | AccInline _ -> None | AccExpr e | AccSet(e,_,_,_) -> Some e.etype) in
+		let e1 = type_access ctx (fst e1) (snd e1) MSet in
+		let e2 = type_expr_with_type ctx e2 (match e1 with AccNo _ | AccInline _ | AccUsing _ -> None | AccExpr e | AccSet(e,_,_,_) -> Some e.etype) in
 		(match e1 with
 		| AccNo s -> error ("Cannot access field or identifier " ^ s ^ " for writing") p
 		| AccExpr e1 ->
@@ -578,17 +659,17 @@ let rec type_binop ctx op e1 e2 p =
 			mk (TBinop (op,e1,e2)) e1.etype p
 		| AccSet (e,m,t,_) ->
 			unify ctx e2.etype t p;
-			mk (TCall (mk (TField (e,m)) (tfun [t] t) p,[e2])) t p
-		| AccInline _ ->
+			make_call ctx (mk (TField (e,m)) (tfun [t] t) p) [e2] t p
+		| AccInline _ | AccUsing _ ->
 			assert false)
 	| OpAssignOp op ->
-		(match type_access ctx (fst e1) (snd e1) false with
+		(match type_access ctx (fst e1) (snd e1) MSet with
 		| AccNo s -> error ("Cannot access field or identifier " ^ s ^ " for writing") p
 		| AccExpr e ->
 			let eop = type_binop ctx op e1 e2 p in
 			(match eop.eexpr with
 			| TBinop (_,_,e2) ->
-				unify ctx e2.etype e.etype p;
+				unify ctx eop.etype e.etype p;
 				check_assign ctx e;
 				mk (TBinop (OpAssignOp op,e,e2)) e.etype p;
 			| _ ->
@@ -602,9 +683,9 @@ let rec type_binop ctx op e1 e2 p =
 			l();
 			mk (TBlock [
 				mk (TVars [v,e.etype,Some e]) ctx.api.tvoid p;
-				mk (TCall (mk (TField (ev,m)) (tfun [t] t) p,[get])) t p
+				make_call ctx (mk (TField (ev,m)) (tfun [t] t) p) [get] t p
 			]) t p
-		| AccInline _ ->
+		| AccInline _ | AccUsing _ ->
 			assert false)
 	| _ ->
 	let e1 = type_expr ctx e1 in
@@ -748,7 +829,7 @@ let rec type_binop ctx op e1 e2 p =
 
 and type_unop ctx op flag e p =
 	let set = (op = Increment || op = Decrement) in
-	let acc = type_access ctx (fst e) (snd e) (not set) in
+	let acc = type_access ctx (fst e) (snd e) (if set then MSet else MGet) in
 	let access e =
 		let t = (match op with
 		| Not ->
@@ -773,10 +854,10 @@ and type_unop ctx op flag e p =
 	in
 	match acc with
 	| AccExpr e -> access e
-	| AccInline _ when not set -> access (acc_get acc p)
+	| AccInline _ | AccUsing _ when not set -> access (acc_get ctx acc p)
 	| AccNo s ->
 		error ("The field or identifier " ^ s ^ " is not accessible for " ^ (if set then "writing" else "reading")) p
-	| AccInline _ ->
+	| AccInline _ | AccUsing _ ->
 		error "This kind of operation is not supported" p
 	| AccSet (e,m,t,f) ->
 		let l = save_locals ctx in
@@ -792,7 +873,7 @@ and type_unop ctx op flag e p =
 			l();
 			mk (TBlock [
 				mk (TVars [v,e.etype,Some e]) ctx.api.tvoid p;
-				mk (TCall (mk (TField (ev,m)) (tfun [t] t) p,[get])) t p
+				make_call ctx (mk (TField (ev,m)) (tfun [t] t) p) [get] t p
 			]) t p
 		| Postfix ->
 			let v2 = gen_local ctx t in
@@ -803,7 +884,7 @@ and type_unop ctx op flag e p =
 			l();
 			mk (TBlock [
 				mk (TVars [v,e.etype,Some e; v2,t,Some get]) ctx.api.tvoid p;
-				mk (TCall (mk (TField (ev,m)) (tfun [plusone.etype] t) p,[plusone])) t p;
+				make_call ctx (mk (TField (ev,m)) (tfun [plusone.etype] t) p) [plusone] t p;
 				ev2
 			]) t p
 
@@ -818,7 +899,7 @@ and type_switch ctx e cases def need_val p =
 		| (EConst (Ident name),p) :: l
 		| (EConst (Type name),p) :: l ->
 			(try
-				let e = acc_get (type_ident ctx name false p true) p in
+				let e = acc_get ctx (type_ident ctx name false p MGet) p in
 				(match e.eexpr with
 				| TEnumField (e,_) -> Some (e, List.map (fun _ -> mk_mono()) e.e_types)
 				| _ -> None)
@@ -957,17 +1038,17 @@ and type_switch ctx e cases def need_val p =
 		let cases = List.map matchs cases in
 		mk (TMatch (e,(en,enparams),List.map indexes cases,def)) t p
 
-and type_access ctx e p get =
+and type_access ctx e p mode =
 	match e with
 	| EConst (Ident s) ->
-		type_ident ctx s false p get
+		type_ident ctx s false p mode
 	| EConst (Type s) ->
-		type_ident ctx s true p get
+		type_ident ctx s true p mode
 	| EField _
 	| EType _ ->
 		let fields path e =
 			List.fold_left (fun e (f,_,p) ->
-				let e = acc_get (e true) p in
+				let e = acc_get ctx (e MGet) p in
 				type_field ctx e f p
 			) e path
 		in
@@ -1027,7 +1108,7 @@ and type_access ctx e p get =
 			| _ ->
 				fields acc (type_access ctx (fst e) (snd e))
 		in
-		loop [] (e,p) get
+		loop [] (e,p) mode
 	| EArray (e1,e2) ->
 		let e1 = type_expr ctx e1 in
 		let e2 = type_expr ctx e2 in
@@ -1061,7 +1142,7 @@ and type_expr ctx ?(need_val=true) (e,p) =
 	| EArray _
 	| EConst (Ident _)
 	| EConst (Type _) ->
-		acc_get (type_access ctx e p true) p
+		acc_get ctx (type_access ctx e p MGet) p
 	| EConst (Regexp (r,opt)) ->
 		let str = mk (TConst (TString r)) ctx.api.tstring p in
 		let opt = mk (TConst (TString opt)) ctx.api.tstring p in
@@ -1169,11 +1250,11 @@ and type_expr ctx ?(need_val=true) (e,p) =
 						unify_raise ctx e1.etype t e1.epos;
 						e1
 					with Error (Unify _,_) ->
-						let acc = acc_get (type_field ctx e1 "iterator" e1.epos true) e1.epos in
+						let acc = acc_get ctx (type_field ctx e1 "iterator" e1.epos MCall) e1.epos in
 						match follow acc.etype with
 						| TFun ([],it) ->
 							unify ctx it t e1.epos;
-							mk (TCall (acc,[])) t e1.epos
+							make_call ctx acc [] t e1.epos
 						| _ ->
 							error "The field iterator is not a method" e1.epos
 					)
@@ -1402,6 +1483,32 @@ and type_expr ctx ?(need_val=true) (e,p) =
 				| _ -> t)
 			| t -> t
 		) in
+		(*
+			add 'using' methods compatible with this type
+		*)
+		let rec loop acc = function
+			| [] -> acc
+			| x :: l -> 
+				let acc = ref (loop acc l) in
+				(match x with
+				| TClassDecl c ->
+					let rec dup t = Type.map dup t in
+					List.iter (fun f ->
+						match follow (field_type f) with
+						| TFun ((_,_,t) :: args, ret) when (try unify_raise ctx (dup e.etype) t e.epos; true with _ -> false) ->
+							let f = { f with cf_type = TFun (args,ret); cf_params = [] } in
+							acc := PMap.add f.cf_name f (!acc)
+						| _ -> ()
+					) c.cl_ordered_statics
+				| _ -> ());
+				!acc
+		in	
+		let use_methods = loop PMap.empty ctx.local_using in
+		let t = (if PMap.is_empty use_methods then t else match follow t with
+			| TFun _ -> t (* don't provide use methods for functions *)
+			| TAnon a -> TAnon { a_fields = PMap.fold (fun f acc -> PMap.add f.cf_name f acc) a.a_fields use_methods; a_status = ref Closed; }
+			| _ -> TAnon { a_fields = use_methods; a_status = ref Closed }
+		) in
 		raise (Display t)
 	| EDisplayNew t ->
 		let t = Typeload.load_normal_type ctx t p true in
@@ -1439,7 +1546,7 @@ and type_call ctx e el p =
 						tf_args = missing_args;
 						tf_type = ret;
 						tf_expr = mk (TReturn (Some (
-							mk (TCall (vexpr fun_arg,List.map vexpr (first_args @ missing_args))) ret p
+							make_call ctx (vexpr fun_arg) (List.map vexpr (first_args @ missing_args)) ret p
 						))) ret p;
 					}) (TFun (fun_args missing_args,ret)) p in
 					let func = mk (TFunction {
@@ -1483,23 +1590,23 @@ and type_call ctx e el p =
 	| _ ->
 		(match e with
 		| EField ((EConst (Ident "super"),_),_) , _ | EType ((EConst (Ident "super"),_),_) , _ -> ctx.super_call <- true
-		| _ -> ());
-		match type_access ctx (fst e) (snd e) true with
+		| _ -> ());		
+		match type_access ctx (fst e) (snd e) MCall with
 		| AccInline (ethis,f,t) ->
 			let params, tret = (match follow t with
 				| TFun (args,r) -> unify_call_params ctx (Some f.cf_name) el args p true, r
 				| _ -> error (s_type (print_context()) t ^ " cannot be called") p
 			) in
-			ignore(follow f.cf_type); (* force evaluation *)
-			(match f.cf_expr with
-			| Some { eexpr = TFunction fd } ->
-				let i = if ctx.doinline then Optimizer.type_inline ctx f fd ethis params tret p else None in
-				(match i with
-				| None -> mk (TCall (mk (TField (ethis,f.cf_name)) t p,params)) tret p
-				| Some e -> e)
-			| _ -> error "Recursive inline is not supported" p)
+			make_call ctx (mk (TField (ethis,f.cf_name)) t p) params tret p
+		| AccUsing (et,eparam) ->
+			let fname = (match et.eexpr with TField (_,f) -> f | _ -> assert false) in
+			let params, tret = (match follow et.etype with
+				| TFun ( _ :: args,r) -> unify_call_params ctx (Some fname) el args p false, r
+				| _ -> assert false
+			) in
+			make_call ctx et (eparam::params) tret p
 		| acc ->
-			let e = acc_get acc p in
+			let e = acc_get ctx acc p in
 			let el , t = (match follow e.etype with
 			| TFun (args,r) ->
 				let el = unify_call_params ctx (match e.eexpr with TField (_,f) -> Some f | _ -> None) el args p false in
@@ -1717,6 +1824,7 @@ let create com =
 		locals_map = PMap.empty;
 		locals_map_inv = PMap.empty;
 		local_types = [];
+		local_using = [];
 		type_params = [];
 		curmethod = "";
 		curclass = null_class;
@@ -1765,3 +1873,6 @@ let create com =
 	| [TClassDecl c] -> ctx.api.tarray <- (fun t -> TInst (c,[t]))
 	| _ -> assert false);
 	ctx
+
+;;
+type_field_rec := type_field
\ No newline at end of file
