@@ -28,6 +28,8 @@ type parse_ctx = {
 	mutable classes : hl_class array;
 	mutable jumps : (int * int) list;
 	mutable pos : int;
+	delta_mt : int;
+	delta_cl : int;
 }
 
 let get = As3parse.iget
@@ -39,12 +41,6 @@ let name ctx n = ctx.names.(idx n)
 let method_type ctx n = ctx.methods.(idx (no_nz n))
 let getclass ctx n = ctx.classes.(idx (no_nz n))
 
-let global_mark = ref 0
-
-let alloc_mark() =
-	incr global_mark;
-	!global_mark
-
 let opt f ctx = function
 	| None -> None
 	| Some x -> Some (f ctx x)
@@ -53,8 +49,10 @@ let stack_delta = function
 	| HBreakPoint -> 0
 	| HNop -> 0
 	| HThrow -> -1
-	| HGetSuper _ -> 1  (* ??? *)
-	| HSetSuper _ -> -1 (* ??? *)
+	| HGetSuper _ -> 0
+	| HSetSuper _ -> -1
+	| HDxNs _ -> 0
+	| HDxNsLate -> -1
 	| HRegKill _ -> 0
 	| HLabel -> 0
 	| HJump (cond,_) ->
@@ -100,10 +98,12 @@ let stack_delta = function
 	| HCallPropLex (_,n) -> -n
 	| HCallSuperVoid (_,n) -> -(n + 1)
 	| HCallPropVoid (_,n) -> -(n + 1)
+	| HApplyType n -> -n
 	| HObject n -> -(n * 2) + 1
 	| HArray n -> -n + 1
 	| HNewBlock -> 1
 	| HClassDef _ -> 0
+	| HGetDescendants _ -> 0
 	| HCatch _ -> 1
 	| HFindPropStrict _ -> 1
 	| HFindProp _ -> 1
@@ -145,6 +145,8 @@ let stack_delta = function
 	| HOp op ->
 		(match op with
 		| A3ONeg | A3OINeg | A3OIncr | A3ODecr | A3ONot | A3OBitNot | A3OIIncr | A3OIDecr -> 0
+		| A3OMemGet8 | A3OMemGet16 | A3OMemGet32 | A3OMemGetFloat | A3OMemGetDouble | A3OSign1 | A3OSign8 | A3OSign16 -> 0
+		| A3OMemSet8 | A3OMemSet16 | A3OMemSet32 | A3OMemSetFloat | A3OMemSetDouble -> -2
 		| _ -> -1)
 	| HUnk _ -> assert false
 
@@ -154,6 +156,8 @@ let parse_opcode ctx i = function
 	| A3Throw -> HThrow
 	| A3GetSuper n -> HGetSuper (name ctx n)
 	| A3SetSuper n -> HSetSuper (name ctx n)
+	| A3DxNs s -> HDxNs (ident ctx s)
+	| A3DxNsLate -> HDxNsLate
 	| A3RegKill r -> HRegKill r
 	| A3Label -> HLabel
 	| A3Jump (j,n) ->
@@ -198,10 +202,12 @@ let parse_opcode ctx i = function
 	| A3CallPropLex (p,n) -> HCallPropLex (name ctx p,n)
 	| A3CallSuperVoid (p,n) -> HCallSuperVoid (name ctx p,n)
 	| A3CallPropVoid (p,n) -> HCallPropVoid (name ctx p,n)
+	| A3ApplyType n -> HApplyType n
 	| A3Object n -> HObject n
 	| A3Array n -> HArray n
 	| A3NewBlock -> HNewBlock
 	| A3ClassDef n -> HClassDef (getclass ctx n)
+	| A3GetDescendants p -> HGetDescendants (name ctx p)
 	| A3Catch n -> HCatch n
 	| A3FindPropStrict p -> HFindPropStrict (name ctx p)
 	| A3FindProp p -> HFindProp (name ctx p)
@@ -257,7 +263,7 @@ let parse_code ctx f trys =
 	let codepos pos delta =
 		let id = (try DynArray.get indexes (pos + delta) with _ -> -1) in
 		if id = -1 then begin
-			Printf.eprintf "MISALIGNED JUMP AT %d %c %d IN #%d\n" pos (if delta < 0 then '-' else '+') (if delta < 0 then -delta else delta) (idx (no_nz f.fun3_id));
+			(*Printf.eprintf "MISALIGNED JUMP AT %d %c %d IN #%d\n" pos (if delta < 0 then '-' else '+') (if delta < 0 then -delta else delta) (idx (no_nz f.fun3_id));*)
 			DynArray.get indexes pos; (* jump 0 *)
 		end else
 			id
@@ -362,7 +368,7 @@ let parse_namespace ctx = function
 
 let parse_nset ctx l = List.map (fun n -> ctx.namespaces.(idx n)) l
 
-let rec parse_name ctx = function
+let rec parse_name names ctx = function
 	| A3MName (id,ns) ->
 		(match ctx.namespaces.(idx ns) with
 		| HNPublic p ->
@@ -374,7 +380,8 @@ let rec parse_name ctx = function
 	| A3MRuntimeName id -> HMRuntimeName (ident ctx id)
 	| A3MRuntimeNameLate -> HMRuntimeNameLate
 	| A3MMultiNameLate ns -> HMMultiNameLate ctx.nsets.(idx ns)
-	| A3MAttrib multi -> HMAttrib (parse_name ctx multi)
+	| A3MAttrib multi -> HMAttrib (parse_name names ctx multi)
+	| A3MParams (id,pl) -> HMParams (parse_name names ctx names.(idx id),List.map (fun id -> parse_name names ctx names.(idx id)) pl)
 
 let parse_try_catch ctx t =
 	{
@@ -397,17 +404,17 @@ let parse_function ctx f =
 			if f.f3_metas <> None then assert false;
 			match f.f3_kind with
 			| A3FVar v ->
-				if v.v3_const then assert false;
 				(* v3_value can be <> None if it's a fun parameter with a default value
 					- which looks like a bug of the AS3 compiler *)
-				name ctx f.f3_name , opt name ctx v.v3_type , f.f3_slot
+				name ctx f.f3_name , opt name ctx v.v3_type , f.f3_slot, v.v3_const
 			| _ -> assert false
 		) f.fun3_locals;
 	}
 
-let parse_method_type ctx m f =
+let parse_method_type ctx idx f =
+	let m = ctx.as3.as3_method_types.(idx) in
 	{
-		hlmt_mark = alloc_mark();
+		hlmt_index = idx + ctx.delta_mt;
 		hlmt_ret = opt name ctx m.mt3_ret;
 		hlmt_args = List.map (opt name ctx) m.mt3_args;
 		hlmt_native = m.mt3_native;
@@ -422,8 +429,9 @@ let parse_method_type ctx m f =
 		hlmt_function = opt parse_function ctx f;
 	}
 
-let parse_class ctx c s =
+let parse_class ctx c s index =
 	{
+		hlc_index = index + ctx.delta_cl;
 		hlc_name = name ctx c.cl3_name;
 		hlc_super = opt name ctx c.cl3_super;
 		hlc_sealed = c.cl3_sealed;
@@ -443,7 +451,7 @@ let parse_static ctx s =
 		hls_fields = Array.map (parse_field ctx) s.st3_fields;
 	}
 
-let parse t =
+let parse ?(delta_mt=0) ?(delta_cl=0) t =
 	let ctx = {
 		as3 = t;
 		namespaces = [||];
@@ -453,17 +461,19 @@ let parse t =
 		classes = [||];
 		jumps = [];
 		pos = 0;
+		delta_mt = delta_mt;
+		delta_cl = delta_cl;
 	} in
 	ctx.namespaces <- Array.map (parse_namespace ctx) t.as3_namespaces;
 	ctx.nsets <- Array.map (parse_nset ctx) t.as3_nsets;
-	ctx.names <- Array.map (parse_name ctx) t.as3_names;
+	ctx.names <- Array.map (parse_name t.as3_names ctx) t.as3_names;
 	let hfunctions = Hashtbl.create 0 in
 	Array.iter (fun f -> Hashtbl.add hfunctions (idx (no_nz f.fun3_id)) f) t.as3_functions;
 	ctx.methods <- Array.mapi (fun i m ->
-		parse_method_type ctx t.as3_method_types.(i) (try Some (Hashtbl.find hfunctions i) with Not_found -> None);
+		parse_method_type ctx i (try Some (Hashtbl.find hfunctions i) with Not_found -> None);
 	) t.as3_method_types;
 	ctx.classes <- Array.mapi (fun i c ->
-		parse_class ctx c t.as3_statics.(i)
+		parse_class ctx c t.as3_statics.(i) i
 	) t.as3_classes;
 	let inits = List.map (parse_static ctx) (Array.to_list t.as3_inits) in
 	Array.iter (fun f ->
@@ -477,14 +487,17 @@ let parse t =
 (*			FLATTEN															*)
 (* ************************************************************************ *)
 
-type ('hl,'item,'key) gen_lookup = {
-	h : ('key,int) Hashtbl.t;
+type ('hl,'item) lookup = {
+	h : ('hl,int) Hashtbl.t;
 	a : 'item DynArray.t;
 	f : flatten_ctx -> 'hl -> 'item;
-	k : 'hl -> 'key;
 }
 
-and ('hl,'item) lookup = ('hl,'item,'hl) gen_lookup
+and ('hl,'item) index_lookup = {
+	ordered_list : 'hl list;
+	ordered_array : 'item option DynArray.t;
+	map_f : flatten_ctx -> 'hl -> 'item;
+}
 
 and flatten_ctx = {
 	fints : (hl_int,as3_int) lookup;
@@ -495,38 +508,58 @@ and flatten_ctx = {
 	fnsets : (hl_ns_set,as3_ns_set) lookup;
 	fnames : (hl_name,as3_multi_name) lookup;
 	fmetas : (hl_metadata,as3_metadata) lookup;
-	fmethods : (hl_method,as3_method_type,int) gen_lookup;
-	fclasses : (hl_class,as3_class * as3_static,hl_name) gen_lookup;
+	fmethods : (hl_method,as3_method_type) index_lookup;
+	fclasses : (hl_class,as3_class * as3_static) index_lookup;
 	mutable ffunctions : as3_function list;
 	mutable fjumps : int list;
 }
 
-let new_gen_lookup f k =
+let new_lookup f =
 	{
 		h = Hashtbl.create 0;
 		a = DynArray.create();
 		f = f;
-		k = k;
 	}
 
-let new_lookup f = new_gen_lookup f (fun x -> x)
+let new_index_lookup l f =
+	{
+		ordered_list = l;
+		ordered_array = DynArray.init (List.length l) (fun _ -> None);
+		map_f = f;
+	}
 
 let lookup_array l = DynArray.to_array l.a
 
-let lookup ctx (l:('a,'b,'k) gen_lookup) item : 'b index =
-	let key = l.k item in
+let lookup_index_array l =
+	Array.map (function None -> assert false | Some x -> x) (DynArray.to_array l.ordered_array)
+
+let lookup ctx (l:('a,'b) lookup) item : 'b index =
 	let idx = try
-		Hashtbl.find l.h key
+		Hashtbl.find l.h item
 	with Not_found ->
 		let idx = DynArray.length l.a in
 		(* set dummy value for recursion *)
 		DynArray.add l.a (Obj.magic 0);
-		Hashtbl.add l.h key (idx + 1);
+		Hashtbl.add l.h item (idx + 1);
 		DynArray.set l.a idx (l.f ctx item);
 		idx + 1
 	in
 	As3parse.magic_index idx
 
+let lookup_index_nz ctx (l:('a,'b) index_lookup) item : 'c index_nz =
+	let rec loop n = function
+		| [] -> assert false
+		| x :: l ->
+			if x == item then n else loop (n + 1) l
+	in
+	let idx = loop 0 l.ordered_list in
+	if DynArray.get l.ordered_array idx = None then begin
+		(* set dummy value for recursion *)
+		DynArray.set l.ordered_array idx (Some (Obj.magic 0));
+		DynArray.set l.ordered_array idx (Some (l.map_f ctx item));
+	end;
+	As3parse.magic_index_nz idx
+
 let lookup_nz ctx l item =
 	As3parse.magic_index_nz (As3parse.index_int (lookup ctx l item) - 1)
 
@@ -535,10 +568,10 @@ let lookup_ident ctx i = lookup ctx ctx.fidents i
 let lookup_name ctx n = lookup ctx ctx.fnames n
 
 let lookup_method ctx m : as3_method_type index_nz =
-	lookup_nz ctx ctx.fmethods m
+	lookup_index_nz ctx ctx.fmethods m
 
 let lookup_class ctx c : as3_class index_nz =
-	lookup_nz ctx ctx.fclasses c
+	lookup_index_nz ctx ctx.fclasses c
 
 let flatten_namespace ctx = function
 	| HNPrivate i -> A3NPrivate (opt lookup_ident ctx i)
@@ -562,6 +595,7 @@ let rec flatten_name ctx = function
 	| HMRuntimeNameLate -> A3MRuntimeNameLate
 	| HMMultiNameLate ns -> A3MMultiNameLate (lookup ctx ctx.fnsets ns)
 	| HMAttrib n -> A3MAttrib (flatten_name ctx n)
+	| HMParams (i,nl) -> A3MParams (lookup_name ctx i,List.map (lookup_name ctx) nl)
 
 let flatten_meta ctx m =
 	{
@@ -628,6 +662,8 @@ let flatten_opcode ctx i = function
 	| HThrow -> A3Throw
 	| HGetSuper n -> A3GetSuper (lookup_name ctx n)
 	| HSetSuper n -> A3SetSuper (lookup_name ctx n)
+	| HDxNs s -> A3DxNs (lookup_ident ctx s)
+	| HDxNsLate -> A3DxNsLate
 	| HRegKill r -> A3RegKill r
 	| HLabel -> A3Label
 	| HJump (j,n) ->
@@ -672,10 +708,12 @@ let flatten_opcode ctx i = function
 	| HCallPropLex (i,n) -> A3CallPropLex (lookup_name ctx i,n)
 	| HCallSuperVoid (i,n) -> A3CallSuperVoid (lookup_name ctx i,n)
 	| HCallPropVoid (i,n)-> A3CallPropVoid (lookup_name ctx i,n)
+	| HApplyType n -> A3ApplyType n
 	| HObject n -> A3Object n
 	| HArray n -> A3Array n
 	| HNewBlock -> A3NewBlock
 	| HClassDef c -> A3ClassDef (As3parse.magic_index_nz (As3parse.index_nz_int (lookup_class ctx c)))
+	| HGetDescendants i -> A3GetDescendants (lookup_name ctx i)
 	| HCatch n -> A3Catch n
 	| HFindPropStrict i -> A3FindPropStrict (lookup_name ctx i)
 	| HFindProp i -> A3FindProp (lookup_name ctx i)
@@ -738,7 +776,7 @@ let flatten_code ctx hcode trys =
 		Array.set code j (match Array.get code j with
 			| A3Jump (jc,n) ->
 				A3Jump (jc,positions.(j+n) - positions.(j+1))
-			| A3Switch (n,infos) -> 
+			| A3Switch (n,infos) ->
 				A3Switch (positions.(j+n) - positions.(j),List.map (fun n -> positions.(j+n) - positions.(j)) infos)
 			| _ -> assert false);
 	) ctx.fjumps;
@@ -765,11 +803,11 @@ let flatten_function ctx f mid =
 		fun3_max_scope = f.hlf_max_scope;
 		fun3_code = code;
 		fun3_trys = trys;
-		fun3_locals = Array.map (fun (n,t,s) ->
+		fun3_locals = Array.map (fun (n,t,s,c) ->
 			{
 				f3_name = lookup_name ctx n;
 				f3_slot = s;
-				f3_kind = A3FVar { v3_type = opt lookup_name ctx t; v3_value = A3VNone; v3_const = false  };
+				f3_kind = A3FVar { v3_type = opt lookup_name ctx t; v3_value = A3VNone; v3_const = c };
 				f3_metas = None;
 			}
 		) f.hlf_locals;
@@ -800,8 +838,49 @@ let flatten_static ctx s =
 		st3_fields = Array.map (flatten_field ctx) s.hls_fields;
 	}
 
+let rec browse_method ctx m =
+	let ml, _ = ctx in
+	if not (List.memq m !ml) then begin
+		ml := m :: !ml;
+		match m.hlmt_function with
+		| None -> ()
+		| Some f ->
+			Array.iter (function
+				| HFunction f | HCallStatic (f,_) -> browse_method ctx f
+				| HClassDef _ -> () (* ignore, should be in fields list anyway *)
+				| _ -> ()
+			) f.hlf_code
+	end
+
+and browse_class ctx c =
+	let _, cl = ctx in
+	if not (List.memq c !cl) then begin
+		cl := c :: !cl;
+		browse_method ctx c.hlc_construct;
+		browse_method ctx c.hlc_static_construct;
+		Array.iter (browse_field ctx) c.hlc_fields;
+		Array.iter (browse_field ctx) c.hlc_static_fields;
+	end
+
+and browse_field ctx f =
+	match f.hlf_kind with
+	| HFMethod m -> browse_method ctx m.hlm_type
+	| HFVar _ -> ()
+	| HFFunction m -> browse_method ctx m
+	| HFClass c -> browse_class ctx c
+
 let flatten t =
 	let id _ x = x in
+	(* collect methods and classes, sort by index and force evaluation in order to keep order *)
+	let methods = ref [] in
+	let classes = ref [] in
+	let ctx = (methods,classes) in
+	List.iter (fun s ->
+		Array.iter (browse_field ctx) s.hls_fields;
+		browse_method ctx s.hls_method;
+	) t;
+	let methods = List.sort (fun m1 m2 -> m1.hlmt_index - m2.hlmt_index) (List.rev !methods) in
+	(* done *)
 	let rec ctx = {
 		fints = new_lookup id;
 		fuints = new_lookup id;
@@ -811,13 +890,14 @@ let flatten t =
 		fnsets = new_lookup flatten_ns_set;
 		fnames = new_lookup flatten_name;
 		fmetas = new_lookup flatten_meta;
-		fmethods = new_gen_lookup flatten_method (fun m -> m.hlmt_mark);
-		fclasses = new_gen_lookup flatten_class (fun c -> c.hlc_name);
+		fmethods = new_index_lookup methods flatten_method;
+		fclasses = new_index_lookup (List.rev !classes) flatten_class;
 		fjumps = [];
 		ffunctions = [];
 	} in
 	ignore(lookup_ident ctx "");
 	let inits = List.map (flatten_static ctx) t in
+	let classes = lookup_index_array ctx.fclasses in
 	{
 		as3_ints = lookup_array ctx.fints;
 		as3_uints = lookup_array ctx.fuints;
@@ -827,9 +907,9 @@ let flatten t =
 		as3_nsets = lookup_array ctx.fnsets;
 		as3_names = lookup_array ctx.fnames;
 		as3_metadatas = lookup_array ctx.fmetas;
-		as3_method_types = lookup_array ctx.fmethods;
-		as3_classes = Array.map fst (lookup_array ctx.fclasses);
-		as3_statics = Array.map snd (lookup_array ctx.fclasses);
+		as3_method_types = lookup_index_array ctx.fmethods;
+		as3_classes = Array.map fst classes;
+		as3_statics = Array.map snd classes;
 		as3_functions = Array.of_list (List.rev ctx.ffunctions);
 		as3_inits = Array.of_list inits;
 		as3_unknown = "";
